#!/usr/bin/env python

from numpy import nan, repeat, concatenate, array, unique, reshape
from pandas import Grouper, DataFrame, read_csv, to_datetime
from tqdm import tqdm
from datetime import timedelta

def remove_outlier_from_time_average(df, time=4, multiplier=3):
    """
    
    Remove outliers when averaging transients before performing the fitting routines, used to improve the signal to noise ratio in low biomass systems.

    The function sets a time window to average over, using upper and lower limits for outlier detection.
    The upper and lower limits are determined by mean ± std * [1].
    The multiplier [1] can be adjusted by the user.

    Parameters
    ----------
    df : pandas.DataFrame
        A dataframe of the raw data, can either be imported from pandas.read_csv or the output from phyto_photo_utils.load
    time : int, default=4
     	The time window to average over, e.g. 4 = 4 minute averages
    multiplier : int, default=3	
        The multiplier to apply to the standard deviation for determining the upper and lower limits.

    Returns
    -------
    df : pandas.DataFrame 
        A dataframe of the time averaged data with outliers excluded.

    Example
    -------
    >>> ppu.remove_outlier_from_time_average(df, time=2, multiplier=3)

    """

    # Convert time window to string
    dt = str(time)+'T'
    # Convert dtype of the datetime column
    df['datetime'] = df.datetime.astype('datetime64')
    # Group data by time window and flashlet number
    grp = df.groupby([Grouper(key='datetime', freq=dt), 'flashlet_number'])
    # Calculate means, standard deviations and counts of the groups
    mean = grp.mean()
    std = grp.std()
    c = grp.count()
    # Calculate upper and lower limits of each group, and repeat each value by its count
    ulim = repeat((mean.fyield.values + std.fyield.values * multiplier), c.fyield.values)
    llim = repeat((mean.fyield.values - std.fyield.values * multiplier), c.fyield.values)

    # Get indexes of data used to create each group
    idx = []
    for i, items in enumerate(grp.indices.items()):
        idx.append(items[-1])

    idx = concatenate(idx, axis=0)

    # Create pandas DataFrame of upper and lower using original indexes of data
    mask = DataFrame([ulim, llim, idx]).T
    mask.columns = ['ulim','llim','index']
    mask = mask.set_index('index').sort_index()

    # Create boolean array using mask DataFrame
    m = (df.fyield.values > mask.ulim) | (df.fyield.values < mask.llim)
    
    # Where condition is True, set values of fluorescence yield to NaN
    df.loc[m.values,'fyield'] = nan

    # Group data that is now corrected
    df = df.groupby([Grouper(key='datetime', freq=dt), 'flashlet_number']).mean().reset_index()
    
    # Return number of measurements that is used to create each average
    df['nseq'] = c.fyield.values
    
    return df

def correct_fire_instrument_bias(df, pos=1, sat_len=100):
    
    """
    
    Corrects for instrumentation bias in the relaxation phase by calculating difference between flashlet 0 the relaxation phase & flashlet[pos].
    This bias is then added to the relaxation phase.

    Parameters
    ----------
    df : pandas.DataFrame 
        A dataframe of the raw data, can either be imported from pandas.read_csv or the output from phyto_photo_utils.load
    pos : int, default=1
     	The flashlet number after the start of the phase, either saturation or relaxation, to calculate difference between.
    sat_len : int, default=100
     	The length of saturation measurements.

    Returns
    -------
    df : pandas.DataFrame 
        A dataframe of FIRe data corrected for the instrument bias.

    Example
    -------
    >>> ppu.correct_fire_bias_correction(df, pos=1, sat_len=100)

    """

    fyield = array(df.fyield)
    seq = array(df.seq)
    
    ycorr = []
    
    # Loop through data using unique measurement (seq) number
    for s in tqdm(unique(seq)):

        i = seq == s
        y = fyield[i]
        d = y[sat_len - pos] - y[sat_len]
        y[sat_len:] += d
        
        ycorr.append(y)
        
    fyield_corr = array(ycorr)
    fyield_corr = reshape(fyield_corr, (fyield_corr.shape[0] * fyield_corr.shape[1]))
    
    # Replace fluorescence yield in DataFrame with bias corrected data 
    df['fyield'] = fyield_corr
    
    return df

def calculate_blank_FastOcean(file_, seq_len=100):
     
    """
    Calculates the blank by averaging the fluorescence yield for the saturation phase.

    Parameters
    ----------
    file_ : str
        The path directory to the raw blank file in csv format.
    seq_len : int, default=100
        The length of the measurement sequence.
    
    Returns
    -------
    res : pandas.DataFrame
        The blank results.

    Example
    -------
    >>> ppu.calculate_blank_FastOcean(file_, seq_len=100)
    """

    df = read_csv(file_, skiprows=26, nrows=2, header=None)
    df = df.iloc[:,2:].T
    df.columns = ['date', 'time']
    df['datetime'] = to_datetime(df.date.values+' '+df.time.values)
    df = df.drop(columns=['date','time'])

    res = read_csv(file_, skiprows=43, nrows=seq_len, header=None)
    res = res.iloc[:,2:]
    res = res.agg(['mean','std']).T
    res.columns = ['blank_mean', 'blank_stdev']
    res = DataFrame(res)
    res['datetime'] = df.datetime

    return res

def calculate_blank_FIRe(file_):

    """
    Calculates the blank by averaging the fluorescence yield for the saturation phase.

    Parameters
    ----------
    file_ : str
        The path directory to the raw blank file.

    Returns
    -------
    res : pandas.DataFrame
        The blank results: blank, datetime

    Example
    -------
    >>> ppu.calculate_blank_FIRe(file_)
    """

    df = read_csv(file_)
    # Get date and time data
    dt=str(df.iloc[0,:].values).strip()[2:-2].split('  ')
    date = dt[0].strip()
    time = dt[-1].strip()
    time = str(timedelta(seconds=int(time)))
    datetime = date+' '+time
    format = '%m/%d/%Y %H:%M:%S'
    dt = to_datetime(datetime, format=format)

    # Get saturation phase length
    sat_len = int(str(df.iloc[5,:].values).split()[-1][:-2])

    # Read in the actual data and calculate the mean
    df = read_csv(file_, index_col=0, skiprows=20, header=None, delim_whitespace=True)
    df.columns = ['time', 'ex', 'fyield']
    blank = df.fyield[:sat_len].mean(axis=0)
    stdev = df.fyield[:sat_len].std(axis=0)
    data = array([dt, blank, stdev]).T
    res = DataFrame(data)
    res = res.T
    res.columns = ['datetime', 'blank_mean', 'blank_stdev']

    return res




#!/usr/bin/env python

from numpy import exp, sum, mean, sqrt, diag, linalg, arange, cumsum, nansum, isnan, log10
from sklearn.metrics import mean_squared_error

def __fit_kolber_nop__(pfd, fo, fm, sig):
	return fo + (fm - fo) * (1 - exp(-sig * cumsum(pfd)))

def __calculate_residual_saturation_nop__(p, pfd, fyield):
	return  fyield - __fit_kolber_nop__(pfd, *p)

def __fit_kolber_p__(pfd, fo, fm, sig, ro):
	c = pfd[:] * 0.
	c[0] = pfd[0] * sig
	for i in arange(1, len(pfd)):
		c[i] = c[i-1] + pfd[i] * sig * (1 - c[i-1])/(1 - ro * c[i-1])
	return fo + (fm - fo) * c * (1-ro) / (1-c*ro)

def __calculate_residual_saturation_p__(p, pfd, fyield):
	return  fyield - __fit_kolber_p__(pfd, *p)

def __calculate_residual_saturation_fixedp__(p, pfd, fyield, ro):
	return  fyield - __fit_kolber_p__(pfd, *p, ro)

def __fit_single_relaxation__(seq_time, fo_relax, fm_relax, tau):
	return (fm_relax - (fm_relax - fo_relax) * (1 - exp(-seq_time/tau)))

def __calculate_residual_single_relaxation__(p, seq_time, fyield):
	return fyield - __fit_single_relaxation__(seq_time, *p)

def __fit_triple_relaxation__(seq_time, fo_relax, fm_relax, alpha1, tau1, alpha2, tau2, alpha3, tau3):
	return (fo_relax + (fm_relax - fo_relax) *(alpha1 * 
			exp(-seq_time / tau1) + alpha2 * exp(-seq_time / tau2) + alpha3 * exp(-seq_time / tau3)))

def __calculate_residual_triple_relaxation__(p, seq_time, fyield):
		return fyield - __fit_triple_relaxation__(seq_time, *p)

def __calculate_Webb_model__(E, P, a):
	return P * (1 - exp(-a * E/P))

def __calculate_residual_etr__(p, E, P):
	return P - __calculate_Webb_model__(E, *p)

def __calculate_modified_Webb_model__(E, P, a):
	return P * (1 - exp(-a * E/P)) * (E**-1)

def __calculate_residual_phi__(p, E, P):
	return P - __calculate_modified_Webb_model__(E, *p)

def __calculate_bias__(sol, fyield):
	m = (isnan(sol)) | (isnan(fyield)) | (sol <= 0) | (fyield <= 0)
	return 10 ** ((nansum(log10(sol[~m]) - log10(fyield[~m])) / len(fyield[~m])))

def __calculate_rmse__(res, fyield):
	return sqrt(mean_squared_error(fyield, res+fyield))	

def __calculate_fit_errors__(jac, res):
	pcov = linalg.inv(jac.T.dot(jac)) * mean(res**2)
	return sqrt(diag(pcov))



#!/usr/bin/env python

from pandas import read_csv, to_datetime, DataFrame, concat, Series
from numpy import array, repeat, arange, squeeze, reshape, r_, nansum
from os import chdir
from csv import reader

def load_FIRe_files(file_, append=False, save_files=False, res_path=None, 
                   seq_len=160, flen=1e-6, irrad=None):
    """

    Process the raw data file (.000 format) and convert to a csv with standard formatting.

    Parameters
    ----------
    file_ : str
        The path directory to the data file from benchtop SAtlantic FIRe.
    append : bool, default=False
        If True, multiple files will be concatenated together.
    save_files : bool, default=False
        If True, files will be saved as .csv.
    res_path : str, default=None           
        The path directory where to save files, only required if save_files = True.
    seq_len : int , default=160               
        The number of flashlets in the protocol.
    flen : float, default=1e-6
        The flashlet length in seconds.
    irrad : int, default=None                 
        The LED output in μE m\ :sup:`2` s\ :sup:`-1`.

    Returns
    -------
    df : pandas.DataFrame, shape=[n,6]
        A dataframe of the raw fluorescence data with columns as below:
    flashlet_number : np.array, dtype=int, shape=[n,]
        A sequential number from 1 to ``seq_len``
    flevel : np.array, dtype=float, shape=[n,]
        The raw fluorescence data.
    datetime : np.array, dtype=datetime64, shape=[n,]
        The date and time of measurement.
    seq : np.array, dtype=int, shape=[n,]
        The sample measurement number.
    seq_time : np.array, dtype=float, shape=[n,]
        The measurement sequence time in μs.
    pfd : np.array, dype=float, shape=[n,]
        The photon flux density in μmol photons m\ :sup:`2` s\ :sup:`-1`.

    Example
    -------
    >>> fname = './data/raw/instrument/fire/FIRe_example.000'
    >>> output = './data/raw/ppu/fire/'
    >>> df = ppu.load_FIRe_files(fname, append=False, save_files=True, res_path=output, seq_len=160, flen=1e-6, irrad=47248)
       
    """

    format = '%m/%d/%Y%H:%M:%S.%f'
    
    name = file_.split('/')[-1].split('.')[0]
    data = read_csv(file_, header=None)

    cur_pos = 0
    temp_data = []
    fl_num = []
    final_data = []
    seq_time = []
    index = []

    is_prev_alpha = False
    while True:
        row = data.values[cur_pos][0]
        row_str = row.strip()
        is_alpha = row_str[0].isalpha()

        if not is_prev_alpha and is_alpha:
            row_str = row_str[:-10]
            row_str = row_str.replace(" ","")
            dt_str = row_str[10:]
            dt = to_datetime(dt_str, format=format)

        if len(temp_data) > 0 and (not is_prev_alpha) and is_alpha:
            temp_data = array(temp_data)
            
            fl_num.append(temp_data[:seq_len, 0])
            seq_time.append(temp_data[:seq_len, 1])
            final_data.append(temp_data[:seq_len, 2])
            index.append(dt)
                         
            temp_data = []

        if not is_alpha:
            temp_data.append(list(map(float, row_str.split())))

        cur_pos += 1
        is_prev_alpha = is_alpha

        if cur_pos == len(data.values)-1:
            break
    
    flashlet = array(fl_num)
    flashlet = Series(squeeze(reshape(flashlet, (1, (flashlet.shape[0]*flashlet.shape[1])))))
    time = array(seq_time)
    time = Series(squeeze(reshape(time, (1, (time.shape[0]*time.shape[1])))))
    data = array(final_data)
    datetime = repeat(index, data.shape[1])
    data = Series(squeeze(reshape(data, (1, (data.shape[0]*data.shape[1])))))
    
    df = DataFrame(data=[flashlet, data, datetime, time])
    df = df.T
    df.columns = ['flashlet_number', 'fyield', 'datetime', 'seq_time']
    df['seq'] = repeat(arange(0,(int(df.shape[0] / seq_len)),1), seq_len)
    sigscale = 1e-20
    df['pfd'] = (irrad * 1e-6 * 6.02e23) * flen * sigscale
    df['flevel'] = df.fyield.astype('float')
    df['seq_time'] = df.seq_time.astype('int')
    
    df = df[['flashlet_number','flevel','datetime','seq','seq_time','pfd']]

    list_ = []

    # If True, multiple files will be concatenated together
    if append:
        list_.append(df)
        df = concat(list_, axis=0)
        df['seq'] = repeat(arange(0, (df.shape[0] / seq_len), 1), seq_len)
    
    # If True, files will be saved as .csv in path directory specified by res_path
    if save_files:
        if res_path is None:
            print('WARNING: Files not saved. No output directory specified.')
        else:
            df.to_csv(res_path+str(name)+'.csv')
        
    return df


def load_FASTTrackaI_files(file_, append=False, save_files=False, res_path=None, seq_len=120, irrad=None):
    """

    Process the raw data file and convert to a csv with standard formatting.

    Parameters
    ----------
    file_ : str
        The path directory to the .000 data file from benchtop SAtlantic FIRe.
    append : bool, default=False
        If True, multiple files will be concatenated together.
    save_files : bool, default=False
        If True, files will be saved as .csv.
    res_path : dir, default=None             
        The path directory where to save files, only required if save_files = True.
    seq_len : int, default=120         
        The number of flashlets in the protocol.
    irrad : int, default=None            
        The light/dark chamber photons per count from the calibration file.

    Returns
    -------
    df : pandas.DataFrame, shape=[n,8]
        A dataframe of the raw fluorescence data with columns as below:
    flashlet_number : np.array, dtype=int, shape=[n,]
        A sequential number from 1 to ``seq_len``
    flevel : np.array, dtype=float, shape=[n,]
        The raw fluorescence data.
    datetime : np.array, dtype=datetime64, shape=[n,]
        The date and time of measurement.
    seq : np.array, dtype=int, shape=[n,]
        The sample measurement number.
    seq_time : np.array, dtype=float, shape=[n,]
        The measurement sequence time in μs.
    pfd : np.array, dype=float, shape=[n,]
        The photon flux density in μmol photons m\ :sup:`2` s\ :sup:`-1`.
    channel : np.array, dtype=str, shape=[n,]
        The chamber used for measurements, A = light chamber, B = dark chamber.
    gain : np.array, dtype=int, shape=[n,]
        The gain settings of the instrument.
    
    Example
    -------
    >>> fname = './data/raw/instrument/fasttrackai/FASTTrackaI_example.csv'
    >>> output = './data/raw/ppu/fasttrackai/'
    >>> df = ppu.load_FASTTrackaI_files(fname, append=False, save_files=True, res_path=output, seq_len=120, irrad=545.62e10)
       
    """    
    df = read_csv(file_, header=0)
    name = file_.split('/')[-1].split('.')[0]
    
    # Selecting only rows that contain metadata
    idx = df.shape[0]
    rows = arange(0, idx+1, seq_len+1)

    DESIRED_ROWS = rows
    with open(file_) as input_file:
        creader = reader(input_file)

        desired_rows = [row for row_number, row in enumerate(creader)
                        if row_number in DESIRED_ROWS]

    md = DataFrame(desired_rows)
    # Calculation of date + time for each measurement
    dt = md.iloc[:, 10].str.replace(':','/') + ' ' + md.iloc[:, 11]
    dt = to_datetime(dt.values)
    gain = md.iloc[:, 12].values
    md = md.iloc[:, 1:-18]
    md.columns = ['seq','channel','sflen','sat_len','sper','rflen','rel_len','rper']
    md['gain'] = gain
    md['datetime'] = dt
    
    # Calculation of saturation and relaxation flashlet length
    md['sflen'] = 1.1 + (md.sflen.astype('float64') - 4) * 0.06
    md['rflen'] = 1.1 + (md.rflen.astype('float64') - 4) * 0.06
    md['sper'] = 2.8 + md.sper.astype('float64') * 0.08
    md['rper'] = 2.8 + md.rper.astype('float64') * 0.08
    md['sat_len'] = md.sat_len.astype('float64')
    md['sflen'] = md.sflen.astype('float64')
    md['rel_len'] = md.rel_len.astype('float64')
    md['rflen'] = md.rflen.astype('float64')

    # Calculation of sequence time
    stime = []
    for i in range(md.shape[0]):
        s = (arange(0, md.sat_len[i], 1) * md.sper[i] + md.sflen[i]) * 1e-6
        r = (arange(0, md.rel_len[i], 1) * md.rper[i] + md.rflen[i]) * 1e-6
        r = (r+(md.sper[i]*1e-6))+s[-1]
        y = Series(r_[s,r])
        stime.append(y)
    stime = concat(stime, axis=0)
    
    # Selecting rows that contain the raw data
    df = read_csv(file_, header=None, skiprows=rows)
    df.columns= ['flashlet_number','ex', 'em']
    df['flevel'] = df.em / df.ex
    df['datetime'] = repeat(md.datetime.values, seq_len)
    df['seq'] = repeat(md.seq.values, seq_len)
    df['seq_time'] = stime.values * 1e6
    df['channel'] = repeat(md.channel.values, seq_len)

    # Calculation of photon flux density
    sigscale = 1e-20
    df['pfd'] = (df.ex * irrad) * sigscale
    df['gain'] = repeat(md.gain.values, seq_len) # won't be need if gain correction implemented here
    df = df.drop(['ex','em','flashlet_number'], axis=1)
    idx = int(df.seq.max()) + 1
    df['flashlet_number'] = array(list(arange(0,seq_len,1)) * idx)
    
    ##TO DO##
    #add if statements to handle the gain settings   
    #if irf:
    #    if df.channel == 'A':
    #        if df.gain == 1:
    #            df.fyield /= irf.

    df = df[['flashlet_number','flevel','datetime','seq','seq_time','pfd','channel','gain']]
    
    list_ = []
    # If True, multiple files will be concatenated together
    if append:
        list_.append(df)
        df = concat(list_, axis=0)
        df['seq'] = repeat(arange(0, (df.shape[0] / seq_len), 1), seq_len)
    
    # If True, files will be saved as .csv in path directory specified by res_path
    if save_files:
        if res_path is None:
            print('WARNING: Files not saved. No output directory specified.')
        else:
            df.to_csv(res_path+str(name)+'.csv')
        
    return df

def load_FastOcean_files(file_, append=False, save_files=False, led_separate=False, res_path=None, 
                   seq_len=140, seq_reps=None, flen=1e-6, FastAct=True, Single_Acq=False):
    """

    Process the raw data file and convert to a csv with standard formatting.

    Parameters
    ----------
    file_ : dir
        The path directory to the .csv data file from the FastOcean with either the FastAct1 or FastAct2 laboratory system.
    append : bool, default=False
        If True, multiple files will be concatenated together. Not applicable if Single_Acq = True.
    save_files : bool, default=False
        If True, files will be saved as .csv.
    led_separate : bool, default=False
        If True, the protocols will be separated dependent upon the LED sequence.
    res_path : dir               
        The path directory where to save files, only required if save_files = True.
    seq_len : int, default=140                 
        The number of flashlets in the protocol.
    seq_reps : int, default=None
        The number of sequences per acquisition or in FastAct2 multiple acquisition files the number of light steps.
    flen : float, default=1e-6
        The flashlet length in seconds.
    FastAct : bool, default=True
        If True, will load data from FastAct1 laboratory system format. If False, will load data from FastAct2 laboratory system format.
    Single_Acq : bool, default=False
        If True, will load a single acquisition data from the FastAct2 laboratory system.

    Returns
    -------

    df : pandas.DataFrame, shape=[n,7]
        A dataframe of the raw fluorescence data with columns as below:
    flashlet_number : np.array, dtype=int, shape=[n,]
        A sequential number from 1 to ``seq_len``
    flevel : np.array, dtype=float, shape=[n,]
        The raw fluorescence data.
    datetime : np.array, dtype=datetime64, shape=[n,]
        The date and time of measurement.
    seq : np.array, dtype=int, shape=[n,]
        The sample measurement number.
    seq_time : np.array, dtype=float, shape=[n,]
        The measurement sequence time in μs.
    pfd : np.array, dype=float, shape=[n,]
        The photon flux density in μmol photons m\ :sup:`2` s\ :sup:`-1`.
    led_sequence : np.array, dtype=int, shape=[n,]
        The LED combination using during the measurement, see example below.

    Example
    -------
    >>> fname = './data/raw/instrument/fire/FastOcean_example.csv'
    >>> output = './data/raw/ppu/fastocean/'
    >>> df = ppu.load_FastOcean_files(fname, append=False, save_files=True, led_separate=False, res_path=output, seq_len=140, flen=1e-6)
    >>> led_sequence == 1, LED 450 nm
    >>> led_sequence == 2, LED 450 nm + LED 530 nm
    >>> led_sequence == 3, LED 450 nm + LED 624 nm
    >>> led_sequence == 4, LED 530 nm + LED 624 nm + LED 624 nm
    >>> led_sequence == 5, LED 530 nm + LED 624 nm 
    >>> led_sequence == 6, LED 530 nm
    >>> led_sequence == 7, LED 624 nm
    """    
    name = file_.split('/')[-1].split('.')[0]
    sigscale=1e-20 

    if FastAct:
        md = read_csv(file_, skiprows=16, nrows=12, header=None)
        md = md.T.iloc[2:,:]
        md.columns = ['seq', 'led450', 'led530', 'led624', 'PMT', 'a', 'pitch', 'reps', 'int', 'b', 'date', 'time']
        md = md.drop(['a','b'],axis=1).reset_index(drop=True)
        md['seq'] = md.seq.str.replace('A','').astype('int')
        pfd = md.iloc[:,1:4].astype(float).sum(axis=1)
        nled = md.iloc[:,1:4].count(axis=1)
        
        #determining LED sequence for filtering results
        led = md.iloc[:,1:4].astype(float)
        led = led.isnull()
        led_idx = []
        for i in range(led.shape[0]):
            if (led.iloc[i,1] == True) & (led.iloc[i,2] == True): # LED 1 only
                led_idx.append(1)
            elif (led.iloc[i,0] == False) & (led.iloc[i,1] == False): # LED 1 & 2
                led_idx.append(2)
            elif (led.iloc[i,0] == False) & (led.iloc[i,2] == False): # LED 1 & 3
                led_idx.append(3)
            elif (led.iloc[i,1] == False) & (led.iloc[i,2] == False): # LED 2 & 3
                led_idx.append(5)
            elif (led.iloc[i,0] == False) & (led.iloc[i,1] == False) & (led.iloc[i,2] == False): # ALL LEDs
                led_idx.append(4)
            elif (led.iloc[i,1] == False): # LED 2 only
                led_idx.append(6)
            elif (led.iloc[i,2] == False): # LED 3 only
                led_idx.append(7)
        
        df = read_csv(file_, skiprows=43, nrows=seq_len, header=None)
        seq_time = df.iloc[:,0]
        flashlet_number = df.iloc[:,1]
        df = df.iloc[:,2:]
        
        dfm = []
        for i in range(df.shape[1]):
            data = df.iloc[:,i]
            dfm.append(data)

        df = DataFrame(concat(dfm, axis=0)).reset_index(drop=True)
        df.columns = ['flevel']
        df['pfd'] = repeat((pfd.values*1e22), seq_len) * flen * sigscale
        df['seq'] = repeat(md.seq.values, seq_len)
        df['seq_time'] = array(list(seq_time) * len(md.seq))
        df['flashlet_number'] = array(list(flashlet_number) * len(md.seq))
        df['datetime'] = to_datetime(repeat((md.date.values+' '+md.time.values), seq_len), dayfirst=True)
        df['led_sequence'] = repeat(array(led_idx), seq_len)
        
        df = df[['flashlet_number','flevel','datetime','seq','seq_time','pfd','led_sequence']]
    
    else:
        if Single_Acq:
            df = read_csv(file_, skiprows=2, nrows=18, usecols=[0,1], header=None)
            date = df.iloc[0,1]
            time = df.iloc[1,1]
            datetime = repeat(to_datetime(date+' '+time, dayfirst=True), seq_len * 4 * seq_reps)
            fln = concat([Series(arange(1,seq_len+1, 1))]*4*seq_reps).reset_index(drop=True)
            led_seq = Series(repeat(arange(1,5,1), seq_len*seq_reps))
            led1 = Series(repeat(((float(df.iloc[13,1]) * 1e22) * flen * sigscale), seq_len))
            led2 = Series(repeat(((float(df.iloc[14,1]) * 1e22) * flen * sigscale), seq_len))
            led3 = Series(repeat(((float(df.iloc[15,1]) * 1e22) * flen * sigscale), seq_len))

            df = read_csv(file_, skiprows=41, nrows=seq_len, header=None)
            seq_time = concat([df.iloc[:,1]]*4*seq_reps).reset_index(drop=True)
            df1 = Series(reshape(df.iloc[:,3:(seq_reps+3)].values.T, [-1]))
            df2 = Series(reshape(df.iloc[:,(seq_reps+6):((seq_reps+3)*2)].values.T, [-1]))
            df3 = Series(reshape(df.iloc[:,((seq_reps+4)*2)+1:((seq_reps+3)*3)].values.T, [-1]))
            df4 = Series(reshape(df.iloc[:,((seq_reps+4)*3):(seq_reps*5)-9].values.T, [-1]))
            dfc = concat([df1, df2, df3, df4]).reset_index(drop=True)
            pfd = concat([led1, led1+led2, led1+led3, led1+led2+led3]*seq_reps).reset_index(drop=True)
            df = DataFrame([fln, dfc, datetime, seq_time, pfd, led_seq]).T
            df.columns = ['flashlet_number','flevel','datetime','seq_time','pfd','led_sequence']
        
        else:
            df = read_csv(file_, skiprows=2, nrows=16, usecols=[0,1], header=None)
            date = df.iloc[0,1]
            time = df.iloc[1,1]
            datetime = repeat(to_datetime(date+' '+time, dayfirst=True), seq_len * 4 * seq_reps)
            fln = concat([Series(arange(1,seq_len+1, 1))] * 4 * seq_reps).values
            led_seq = Series(repeat(arange(1,5,1), seq_len * seq_reps))
            led1 = Series(repeat(((float(df.iloc[13,1]) * 1e22) * flen * sigscale), seq_len))
            led2 = Series(repeat(((float(df.iloc[14,1]) * 1e22) * flen * sigscale), seq_len))
            led3 = Series(repeat(((float(df.iloc[15,1]) * 1e22) * flen * sigscale), seq_len))

            df = read_csv(file_, skiprows=46, nrows=seq_len, header=None)
            seq_time = concat([df.iloc[:,0]] * (4 * seq_reps))
            df1 = Series(reshape(df.iloc[:,2:2+seq_reps].values.T, [-1]))
            df = read_csv(file_, skiprows=212, nrows=seq_len, header=None)
            df2 = Series(reshape(df.iloc[:,2:2+seq_reps].values.T, [-1]))
            df = read_csv(file_, skiprows=378, nrows=seq_len, header=None)
            df3 = Series(reshape(df.iloc[:,2:2+seq_reps].values.T, [-1]))
            df = read_csv(file_, skiprows=544, nrows=seq_len, header=None)
            df4 = Series(reshape(df.iloc[:,2:2+seq_reps].values.T, [-1]))
            dfc = concat([df1, df2, df3, df4]).reset_index(drop=True)
            pfd = repeat(concat([led1, led1+led2, led1+led3, led1+led2+led3]).reset_index(drop=True), seq_reps).values
            df = DataFrame([fln, dfc, datetime, seq_time, pfd, led_seq]).T
            df.columns = ['flashlet_number','flevel','datetime','seq_time','pfd','led_sequence']

    list_ = []
    if append:
            if Single_Acq:
                print('Unable to concatenate files from a single acquisition')
                pass
            else:
                list_.append(df)
                df = concat(list_, axis=0)
                df['seq'] = repeat(arange(0, (df.shape[0] / seq_len), 1), seq_len)
    
    if save_files:
        if res_path is None:
            print('WARNING: Files not saved. No output directory specified.')
        else:
            if Single_Acq:
                df['seq'] = Series(repeat(arange(1,(df.shape[0] / seq_len)+1, 1), seq_len)).values.astype(int)
                df = df[['flashlet_number','flevel','datetime','seq','seq_time','pfd','led_sequence']]
            else:
                df['seq'] = Series(repeat(arange(1,(df.shape[0] / seq_len)+1, 1), seq_len)).values.astype(int)
                df = df[['flashlet_number','flevel','datetime','seq','seq_time','pfd','led_sequence']]
            df.to_csv(res_path+str(name)+'.csv')
            
            if led_separate:
                i = df.led_sequence == 1
                dfi = df[i].reset_index(drop=True)
                if len(dfi) > 0:
                    if Single_Acq:
                        dfi.loc[:,'seq'] = Series(repeat(arange(1, seq_reps+1, 1), seq_len)).values.astype(int)
                        dfi = dfi[['flashlet_number','flevel','datetime','seq','seq_time','pfd','led_sequence']]
                    else:
                        dfi.loc[:,'seq'] = Series(repeat(arange(1, seq_reps+1, 1), seq_len)).values.astype(int)
                        dfi = dfi[['flashlet_number','flevel','datetime','seq','seq_time','pfd','led_sequence']]
                    dfi.to_csv(res_path+str(name)+'_L1.csv')

                i = df.led_sequence == 2
                dfi = df[i].reset_index(drop=True)
                if len(dfi) > 0:
                    if Single_Acq:
                        dfi.loc[:,'seq'] = Series(repeat(arange(1, seq_reps+1, 1), seq_len)).values.astype(int)
                        dfi = dfi[['flashlet_number','flevel','datetime','seq','seq_time','pfd','led_sequence']]
                    else:
                        dfi.loc[:,'seq'] = Series(repeat(arange(1, seq_reps+1, 1), seq_len)).values.astype(int)
                        dfi = dfi[['flashlet_number','flevel','datetime','seq','seq_time','pfd','led_sequence']]
                    dfi.to_csv(res_path+str(name)+'_L12.csv')

                i = df.led_sequence == 3
                dfi = df[i].reset_index(drop=True)
                if len(dfi) > 0:
                    if Single_Acq:
                        dfi.loc[:,'seq'] = Series(repeat(arange(1, seq_reps+1, 1), seq_len)).values.astype(int)
                        dfi = dfi[['flashlet_number','flevel','datetime','seq','seq_time','pfd','led_sequence']]
                    else:
                        dfi.loc[:,'seq'] = Series(repeat(arange(1, seq_reps+1, 1), seq_len)).values.astype(int)
                        dfi = dfi[['flashlet_number','flevel','datetime','seq','seq_time','pfd','led_sequence']]
                    dfi.to_csv(res_path+str(name)+'_L13.csv')

                i = df.led_sequence == 4
                dfi = df[i].reset_index(drop=True)
                if len(dfi) > 0:
                    if Single_Acq:
                        dfi.loc[:,'seq'] = Series(repeat(arange(1, seq_reps+1, 1), seq_len)).values.astype(int)
                        dfi = dfi[['flashlet_number','flevel','datetime','seq','seq_time','pfd','led_sequence']]
                    else:
                        dfi.loc[:,'seq'] = Series(repeat(arange(1, seq_reps+1, 1), seq_len)).values.astype(int)
                        dfi = dfi[['flashlet_number','flevel','datetime','seq','seq_time','pfd','led_sequence']]
                    dfi.to_csv(res_path+str(name)+'_L123.csv')

                i = df.led_sequence == 5
                dfi = df[i].reset_index(drop=True)
                if len(dfi) > 0:
                    if Single_Acq:
                        dfi.loc[:,'seq'] = Series(repeat(arange(1, seq_reps+1, 1), seq_len)).values.astype(int)
                        dfi = dfi[['flashlet_number','flevel','datetime','seq','seq_time','pfd','led_sequence']]
                    else:
                        dfi.loc[:,'seq'] = Series(repeat(arange(1, seq_reps+1, 1), seq_len)).values.astype(int)
                        dfi = dfi[['flashlet_number','flevel','datetime','seq','seq_time','pfd','led_sequence']]
                    dfi.to_csv(res_path+str(name)+'_L23.csv')

                i = df.led_sequence == 6
                dfi = df[i].reset_index(drop=True)
                if len(dfi) > 0:
                    if Single_Acq:
                        dfi.loc[:,'seq'] = Series(repeat(arange(1, seq_reps+1, 1), seq_len)).values.astype(int)
                        dfi = dfi[['flashlet_number','flevel','datetime','seq','seq_time','pfd','led_sequence']]
                    else:
                        dfi.loc[:,'seq'] = Series(repeat(arange(1, seq_reps+1, 1), seq_len)).values.astype(int)
                        dfi = dfi[['flashlet_number','flevel','datetime','seq','seq_time','pfd','led_sequence']]
                    dfi.to_csv(res_path+str(name)+'_L2.csv')

                i = df.led_sequence == 7
                dfi = df[i].reset_index(drop=True)
                if len(dfi) > 0:
                    if Single_Acq:
                        dfi.loc[:,'seq'] = Series(repeat(arange(1, seq_reps+1, 1), seq_len)).values.astype(int)
                        dfi = dfi[['flashlet_number','flevel','datetime','seq','seq_time','pfd','led_sequence']]
                    else:
                        dfi.loc[:,'seq'] = Series(repeat(arange(1, seq_reps+1, 1), seq_len)).values.astype(int)
                        dfi = dfi[['flashlet_number','flevel','datetime','seq','seq_time','pfd','led_sequence']]
                    dfi.to_csv(res_path+str(name)+'_L3.csv')
    
    return df

def load_LIFT_FRR_files(file_, append=False, save_files=False, res_path=None, seq_len=228):
    """

    Process the raw data file from the Soliense LIFT FRR and convert to a csv with standard formatting.

    Parameters
    ----------
    file_ : dir
        The path directory to the .000 data file from benchtop SAtlantic FIRe.
    append : bool, default=False
        If True, multiple files will be concatenated together.
    save_files : bool, default=False
        If True, files will be saved as .csv.
    res_path : dir               
        The path directory where to save files, only required if save_files = True.
    seq_len : int, default=228                 
        The number of flashlets in the protocol.

    Returns
    -------

    df : pandas.DataFrame, shape=[n,7]
        A dataframe of the raw fluorescence data with columns as below:
    flashlet_number : np.array, dtype=int, shape=[n,]
        A sequential number from 1 to ``seq_len``
    flevel : np.array, dtype=float, shape=[n,]
        The raw fluorescence data.
    datetime : np.array, dtype=datetime64, shape=[n,]
        The date and time of measurement.
    seq : np.array, dtype=int, shape=[n,]
        The sample measurement number.
    seq_time : np.array, dtype=float, shape=[n,]
        The measurement sequence time in μs.
    pfd : np.array, dype=float, shape=[n,]
        The photon flux density in μmol photons m\ :sup:`2` s\ :sup:`-1`.

    """

    df = read_csv(file_, header=0)
    md = df[df['----'].str.contains("DateTime:")]
    date = md.iloc[:,1]
    date = date.str.strip().values
    time = md.iloc[:,2]
    time = time.str.strip().values
    datetime = date+' '+time
    datetime = to_datetime(datetime, yearfirst=True)
    to_drop = ['DateTime:', 'PIF:','PARs:','Light:','Lamps:','Gain:','S/N Ratio:','Position:','Cycles:','DataPt','==========','----']
    df = df.iloc[:,:4]
    df = df[~df['----'].isin(to_drop)]
    to_drop = ['99']
    df = df[~df['Unnamed: 1'].isin(to_drop)]
    columns = ['flashlet_number','seq_time','ex','em']
    df.columns = columns
    df['flevel'] = df.em.astype(float) / df.ex.astype(float)
    seq_len = 228
    df['datetime'] = repeat(datetime.values, seq_len)
    df['pfd'] = df.ex.astype(float) * 1e-6
    df['seq'] = repeat(arange(1, (df.shape[0]/seq_len)+1, 1), seq_len).astype(int)
    df = df.drop(['ex','em'], axis=1)
    df = df[['flashlet_number','flevel','datetime','seq','seq_time','pfd']]
    df = df.reset_index(drop=True)

    list_ = []
    if append:
        list_.append(df)
        df = concat(list_, axis=0)
        df['seq'] = repeat(arange(0, (df.shape[0] / seq_len), 1), seq_len)
    
    if save_files:
        if res_path is None:
            print('WARNING: Files not saved. No output directory specified.')
        else:
            df.to_csv(res_path+str(name)+'.csv')

    return df



#!/usr/bin/env python

from numpy import exp, argmin, abs, nanmin, nanmax, nansum, array
from math import pi
from pandas import read_csv

def calculate_chl_specific_absorption(aptot, blank, ap_lambda, depig=None, chl=None, vol=None, betac1=None, betac2=None, diam=None, bricaud_slope=False, phycobilin=False, norm_750=False):
	"""

	Process the raw absorbance data to produce chlorophyll specific phytoplankton absorption.

	Parameters
	----------

	aptot : np.array, dtype=float, shape=[n,]
		The raw absorbance data.
	blank : np.array, dtype=float, shape=[n,]
	 	The blank absorbance data.
	ap_lambda : np.array, dtype=float, shape=[n,]
		The wavelengths corresponding to the measurements.
	depig : np.array, dtype=float, shape=[n,] 
		The raw depigmented absorbance data.
	chl : float
		The chlorophyll concentration associated with the measurement.
	vol : int
		The volume of water filtered in mL.
	betac1 : int
		The pathlength amplification coefficient 1 (see Stramski et al. 2015). For transmittance mode, 0.679, for transmittance-reflectance mode, 0.719, and for integrating sphere mode, 0.323.
	betac2 : int
		The pathlength amplification coefficient 2 (see Stramski et al. 2015). For transmittance mode, 1.2804, for transmittance-reflectance mode, 1.2287, and for integrating sphere mode, 1.0867.
	diam : float			
		The diameter of filtrate in mm.
	bricaud_slope: bool, default=True
	 	If True, will theoretically calculate detrital slope (see Bricaud & Stramski 1990). If False, will subtract depigmented absorption from total absorption.
	phycobilin: bool, default=False
		If True, will account for high absorption in the green wavelengths (580 - 600 nm) by phycobilin proteins when performing the bricaud_slope detrital correction.
	norm_750: bool, default=False
		If True, will normalise the data to the value at 750 nm.



	Returns
	-------

	aphy : np.array, dtype=float, shape=[n,]
		The chlorophyll specific phytoplankton absorption data.

	Example
	-------
	>>> aphy = ppu.calculate_chl_specific_absorption(pa_data, blank, wavelength, chl=0.19, vol=2000, betac1=0.323, betac2=1.0867, diam=15, bricaud_slope=True, phycobilin=True, norm750=False)
	   
	"""
	aptot = array(aptot)
	ap_lambda = array(ap_lambda)

	aptot -= blank # subtract blank from data

	# Convert from absorbance to absorption
	diam = pi * ((diam / 2000)**2) 
	divol = ((vol / 1e6) / diam) 
	aptot *= 2.303 # conversion from log10 to loge
	aptot /= divol
	aptot = betac1 * aptot**betac2

	# Normalise to minimum value (~750 nm)
	if norm_750:
		dmin = argmin(abs(ap_lambda - 750))
		dmin = aptot[dmin]
		aptot -= dmin

	if bricaud_slope: # See Bricaud & Stramski 1990 for more details
	    
	    # Find unique indices based upon wavelengths measured
		idx380 = argmin(abs(ap_lambda - 380))
		idx505 = argmin(abs(ap_lambda - 505))
		idx580 = argmin(abs(ap_lambda - 580))
		idx600 = argmin(abs(ap_lambda - 600))
		idx692 = argmin(abs(ap_lambda - 692))
		idx750 = argmin(abs(ap_lambda - 750))
	    
		ap750 = aptot[idx750]
		R1 = (0.99 * aptot[idx380]) - aptot[idx505]
	    
		if phycobilin:

			phyco = aptot[idx580:idx600]
			ratio = []

			for i in range(len(phyco)):

				r = phyco.values[i] / aptot[idx692]
				ratio.append(r)

			val, idx = min((val, idx) for (idx, val) in enumerate(ratio))
			idx = phyco.index[idx]
			widx = ap_lambda[idx]
			R2 = (aptot[idx]) - 0.92 * aptot[idx692]

		else:
			R2 = (aptot[idx580]) - 0.92 * aptot[idx692]

		if (R1 <= 0) | (R2 <= 0):
			S = 0

		else:
			R = R1/R2
			S = 0.0001
			L1 = (0.99 * exp(-380 * S)) - exp(-505 * S)

			if phycobilin:
				L2 = exp(-widx * S) - 0.92 * exp(-widx * S)
			else:
				L2 = exp(-580 * S) - 0.92 * exp(-692 * S)
			L = L1/L2

			while (S < 0.03):
				S += 0.0001
				L = (0.99 * exp(-380 * S) - 0.99 * exp(-505 * S)) 
				if phycobilin:
					L = L / (exp(-widx * S) - 0.92 * exp(-692 * S))
				else:
					L = L / (exp(-580 * S) - 0.92 * exp(-692 * S))

				if (L/R) >= 1:
					break
			else: 
				S = 0
		if (S == 0 or S == 0.03):
			A = 0 
		else:
			A = (0.99 * aptot[idx380] - aptot[idx505]) / (0.99 * exp(-380 * S) - exp(-505 * S)) 

		slope = A*exp(-S*ap_lambda) - A*exp(-750*S)

	    # Calculate phytoplankton specific absorption
		aphy = aptot - slope

	else:
		if depig is None:
			print('UserError - no depigmented data provided.')
		# Convert from absorbance to absorption
		depig -= blank
		depig *= 2.303 # Conversion from log10 to loge
		depig /= divol

 		# Normalise to minimum value (~750 nm)
		if norm_750:
			
			depmin = argmin(abs(ap_lambda - 750))
			depmin = depig[depmin]
			depig -= depmin
	    
	    # Calculate phytoplankton specific absorption
		aphy = aptot - depig
	
	if chl is None:

		return aphy
	
	else:
		aphy /= chl

		return aphy
	

def calculate_instrument_led_correction(aphy, ap_lambda, method=None, chl=None, e_background=None, e_insitu=None, e_actinic=None, depth=None, e_led=None, wl=None, constants=None):
	"""

	Calculate the spectral correction factor
	TO DO: Create method to convert all arrays to same length and resolution
	TO DO: Add in functionality to calculate mixed excitation wavelength spectra for FastOcean when using 1+ wavelength

	Parameters
	----------

	aphy : np.array, dtype=float, shape=[n,]
		The wavelength specific phytoplankton absorption coefficients.
	ap_lambda : np.array, dtype=int, shape=[n,]
		The wavelengths associated with the aphy and aphy_star.
	method : 'sigma', 'actinic'
		Choose spectral correction method to either correct SigmaPSII or correct the background actinic light.
	e_background : 'insitu', 'actinic'
		For sigma spectral correction factor, select either insitu light (e.g. for underway or insitu measurements) or actinic light (e.g. for fluorescence light curves) as the background light source
	e_insitu : np.array, dtype=int, shape=[n,]
		The in situ irradiance field, if None is passed then will theoretically calculate in situ light field.
	chl : dtype=float
		Chlorophyll concentration for estimation of Kbio for theoretical in situ light field. If None is passed then chl is set to 1 mg/m3.
	e_actinic : 'fastact'
		Actinic light spectrum e.g. Spectra of the Actinic lights within the FastAct illuminating during Fluorescence Light Curves etc. Must defined for 'actinic' method.
	depth : float, default=None
		The depth of the measurement. Must be set if theoretically calculating e_insitu.
	e_led : 'fire','fasttracka_ii', 'fastocean'
		The excitation spectra of the instrument.
	wl : '450nm', '530nm', 624nm', None
		For FastOcean only. Select the excitation wavelength. Future PPU versions will provide option to mix LEDs.

	Returns
	-------

	scf : float
		The spectral correction factor to correct SigmaPSII or actinic background light depending on method.

	Example
	-------
	>>> ppu.calculate_instrument_led_correction(aphy, wavelength, e_led='fire')

	"""
	aphy = array(aphy)
	ap_lambda = array(ap_lambda)

	idx400 = argmin(abs(ap_lambda - 400))
	idx700 = argmin(abs(ap_lambda - 700))+1
	aphy = array(aphy[idx400:idx700])
	aphy = aphy / nanmax(aphy)

	if method is None:
		print('User must select spectral correction method for correcting sigma or correcting actinic light')

	if constants is None:
		df = read_csv('./data/output/spectral_correction_factors/spectral_correction_constants.csv', index_col=0)
	else:
		df = read_csv(constants, index_col=0)

	if method == 'sigma':

		if chl is None:
			print(r'Chlorophyll concentration set to 1 mg m$^-$$^3$')
			chl = 1
		else:
			chl = chl

		if e_background == 'insitu':

			if e_insitu is None:
				if depth is None:
					print('User must define depth for calculating in situ light spectra')
				else:
					kd = df.a_W + df.bb_W + (df.chi * chl ** df.e)
					Ez = df.Ezero * exp(-kd * depth)
					e_background = Ez / max(Ez)

			else:
				e_background = e_insitu / max(e_insitu)

		elif e_background == 'actinic':

			if e_actinic == 'fastact':
				e_actinic = df.fastact.values
				e_background = e_actinic / max(e_actinic)
			else:
				e_background = e_actinic / max(e_actinic)

		if e_led is None:
			print('No instrument selected. Unable to calculate sigma spectral correction factor.')
		elif e_led == 'fire':
		 	e_led = df.fire.values
		 	e_led = e_led / nanmax(e_led)
		elif e_led == 'fasttracka_ii':
		 	e_led = df.fasttracka_ii.values
		 	e_led = e_led / nanmax(e_led)
		elif e_led == 'fastocean':
		 	if wl == None:
		 		print('User must select single excitation wavelength for FastOcean')
		 	elif wl == '450nm':
		 		e_led = df.fastocean_450.values
		 	elif wl == '530nm':
		 		e_led = df.fastocean_530.values
		 	elif wl == '624nm':
		 		e_led = df.fastocean_624.values

		# Perform SCF calculation for sigma
		scf = (nansum(aphy * e_background) * nansum(e_led)) / (nansum(aphy * e_led) * nansum(e_background))

	elif method == 'actinic':
		if e_insitu is None:
			if depth is None:
				print('User must define depth for calculating in situ light spectra')
			else:
				kd = df.a_W + df.bb_W + (df.chi * chl ** df.e)
				Ez = df.Ezero * exp(-kd * depth)
				e_insitu = Ez / max(Ez)
		else:
			e_insitu = e_insitu / max(e_insitu)

		if e_actinic == 'fastact':
			e_actinic = df.fastact.values
			e_actinic = e_actinic / max(e_actinic)
		else:
			e_actinic = e_actinic / max(e_actinic)

		# Perform SCF calculation for the actinic light

		scf = (nansum(aphy * e_actinic) * nansum(e_insitu)) / (nansum(aphy * e_insitu) * nansum(e_actinic))


	return scf





#!/usr/bin/env python

from ._load import load_FASTTrackaI_files, load_FastOcean_files, load_FIRe_files, load_LIFT_FRR_files
from ._saturation import fit_saturation
from ._relaxation import fit_relaxation
from ._tools import remove_outlier_from_time_average, correct_fire_instrument_bias, calculate_blank_FastOcean, calculate_blank_FIRe
from ._etr import calculate_etr
from ._spectral_correction import calculate_chl_specific_absorption, calculate_instrument_led_correction
from ._equations import __fit_kolber_p__, __fit_kolber_nop__, __calculate_residual_saturation_p__, __calculate_residual_saturation_nop__, __calculate_residual_saturation_fixedp__, __fit_single_relaxation__, __calculate_residual_single_relaxation__, __fit_triple_relaxation__, __calculate_residual_triple_relaxation__, __calculate_Webb_model__, __calculate_residual_etr__, __calculate_modified_Webb_model__, __calculate_residual_phi__, __calculate_rmse__, __calculate_bias__
from ._fitting import __fit_fixed_p_model__, __fit_calc_p_model__, __fit_no_p_model__, __fit_single_decay__, __fit_triple_decay__
from ._plot import plot_saturation_data, plot_relaxation_data, plot_fluorescence_light_curve



#!/usr/bin/env python

from ._fitting import __fit_single_decay__, __fit_triple_decay__
from numpy import array, unique
from pandas import Series, concat
from tqdm import tqdm

def fit_relaxation(fyield, seq_time, seq, datetime, blank=0, sat_len=100, rel_len=60, sat_flashlets=None, single_decay=False, bounds=True, single_lims=[100,50000], tau1_lims=[100, 800], tau2_lims=[800, 2000], tau3_lims=[2000, 50000], method='trf', loss='soft_l1', f_scale=0.1, max_nfev=None, xtol=1e-9):
	"""

	Process the raw transient data and perform the Kolber et al. 1998 relaxation model.

	Parameters
	----------
	seq_time : np.array, dtype=float, shape=[n,] 
		The sequence time of the flashlets in μs.
	fyield : np.array, dtype=float, shape=[n,] 
		The fluorescence yield of the instrument.
	seq : np.array, dtype=int, shape=[n,] 
		The measurement number.
	datetime : np.array, dtype=datetime64, shape=[n,]
		The date & time of each measurement in the numpy datetime64 format.
	blank : np.array, dtype=float, shape=[n,]
		The blank value, must be the same length as fyield.
	sat_len : int, default=100
		The number of flashlets in the saturation sequence.
	rel_len : int, default=60
		The number of flashlets in the relaxation sequence.
	sat_flashlets : int, default=0
		The number of saturation flashlets to include at the start.
	single_decay : bool, default=False
		If True, will fit a single decay relaxation.
	bounds : bool, default=True
		If True, will set lower and upper limit bounds for the estimation, not suitable for methods 'lm'.
	single_lims : [int, int], default=[100, 50000]
	 	The lower and upper limit bounds for fitting τ, only required if single_decay is True. 
	tau1_lims: [int, int], default=[100, 800]
	 	The lower and upper limit bounds for fitting τ1.
	tau2_lims: [int, int], default=[800, 2000]
	 	The lower and upper limit bounds for fitting τ2.
	tau3_lims: [int, int], default=[2000, 50000]
	 	The lower and upper limit bounds for fitting τ3.
	fit_method : str, default='trf'
		The algorithm to perform minimization. See ``scipy.optimize.least_squares`` documentation for more information on non-linear least squares fitting options.
	loss_method : str, default='soft_l1'
		The loss function to be used. Note: Method ‘lm’ supports only ‘linear’ loss. See ``scipy.optimize.least_squares`` documentation for more information on non-linear least squares fitting options.
	fscale : float, default=0.1
	 	The soft margin value between inlier and outlier residuals. See ``scipy.optimize.least_squares`` documentation for more information on non-linear least squares fitting options.
	max_nfev : int, default=None		
		The number of iterations to perform fitting routine. If None, the value is chosen automatically. See ``scipy.optimize.least_squares`` documentation for more information on non-linear least squares fitting options.
	xtol : float, default=1e-9			
		The tolerance for termination by the change of the independent variables. See ``scipy.optimize.least_squares`` documentation for more information on non-linear least squares fitting options.

	Returns
	-------
	res: pandas.DataFrame
		The results of the fitting routine with columns as below:
	fo_r : np.array, dtype=float, shape=[n,]
		The minimum fluorescence level of relaxation phase.
	fm_r : np.array, dtype=float, shape=[n,]
		The maximum fluorescence level of relaxation phase
	tau : np.array, dtype=float, shape=[n,]
		The rate of QA\ :sup:`-` reoxidation in μs, only returned if single_decay is True.
	alpha1 : np.array, dtype=float, shape=[n,]
		The decay coefficient of τ\ :sub:`1`, only returned if single_decay is False.
	tau1 : np.array, dtype=float, shape=[n,]
		The rate of QA\ :sup:`-` reoxidation in μs, only returned if single_decay is False.
	alpha2 : np.array, dtype=float, shape=[n,]
		The decay coefficient of τ\ :sub:`2`.
	tau2 : np.array, dtype=float, shape=[n,]
		The rate of QB\ :sup:`-` reoxidation in μs, only returned if single_decay is False.
	alpha3 : np.array, dtype=float, shape=[n,]
		The decay coefficient of τ\ :sub:`3`, only returned if single_decay is False.
	tau3 : np.array, dtype=float, shape=[n,]
		The rate of PQ reoxidation in μs, only returned if single_decay is False.
	bias : np.array, dtype=float, shape=[n,]
		The bias of fit.
	rmse : np.array, dtype=float, shape=[n,]
		The root mean squared error of the fit.
	fo_err : np.array, dtype=float, shape=[n,]
		The fit error of Fo_relax.
	fm_err : np.array, dtype=float, shape=[n,]
		The fit error of Fm_relax.
	tau_err : np.array, dtype=float, shape=[n,]
		The fit error of τ, only returned if single_decay is True.
	alpha1_err : np.array, dtype=float, shape=[n,]
		The fit error of α\ :sub:`1`, only returned if single_decay is False.
	tau1_err : np.array, dtype=float, shape=[n,]
		The fit error of τ\ :sub:`1`, only returned if single_decay is False.
	alpha2_err : np.array, dtype=float, shape=[n,]
		The fit error of α\ :sub:`2`, only returned if single_decay is False.
	tau2_err : np.array, dtype=float, shape=[n,]
		The fit error of τ\ :sub:`2`, only returned if single_decay is False.
	alpha3_err : np.array, dtype=float, shape=[n,]
		The fit error of α\ :sub:`3`, only returned if single_decay is False.
	tau3_err : np.array, dtype=float, shape=[n,]
		The fit error of τ\ :sub:`3`, only returned if single_decay is False.
	nfl : np.array, dtype=int, shape=[n,]
		The number of flashlets used for fitting.
	niters : np.array, dype=int, shape=[n,]
		The number of functional evaluations done on the fitting routine.
	flag : np.array, dtype=int, shape=[n,]
		The code associated with the fitting routine success, positive values = SUCCESS, negative values = FAILURE.
		-3 : Unable to calculate parameter errors
		-2 : F\ :sub:`o` Relax > F\ :sub:`m` Relax
		-1 : improper input parameters status returned from MINPACK.
		0 : the maximum number of function evaluations is exceeded.
		1 : gtol termination condition is satisfied.
		2 : ftol termination condition is satisfied.
		3 : xtol termination condition is satisfied.
		4 : Both ftol and xtol termination conditions are satisfied.
	success : np.array, dtype=bool, shape=[n,]
		A boolean array reporting whether fit was successful (TRUE) or if not successful (FALSE)
	datetime : np.array, dtype=datetime64, shape=[n,]
		The date and time associated with the measurement.

	Example
	-------
	>>> rel = ppu.calculate_relaxation(fyield, seq_time, seq, datetime, blank=0, sat_len=100, rel_len=40, single_decay=True, bounds=True, tau_lims=[100, 50000])
	"""

		
	seq_time = array(seq_time)
	fyield = array(fyield)
	seq = array(seq)
	dt = array(datetime)
	if single_decay:
		opts = {'sat_flashlets':sat_flashlets, 'bounds':bounds, 'single_lims':single_lims, 'method':method,'loss':loss, 'f_scale':f_scale, 'max_nfev':max_nfev, 'xtol':xtol} 
	else:
		opts = {'sat_flashlets':sat_flashlets, 'bounds':bounds, 'tau1_lims':tau1_lims, 'tau2_lims':tau2_lims, 'tau3_lims':tau3_lims, 'method':method,'loss':loss, 'f_scale':f_scale, 'max_nfev':max_nfev, 'xtol':xtol} 
	res = []
    
	for s in tqdm(unique(seq)):

		i = seq == s
		x = seq_time[i]
		y = fyield[i]
		x_min = min(x[sat_len:])
		x = x[sat_len-sat_flashlets:sat_len+rel_len] - x_min
		y = y[sat_len-sat_flashlets:sat_len+rel_len]
		if single_decay:
			rel = __fit_single_decay__(x, y, **opts)
		else:
			rel = __fit_triple_decay__(x, y, **opts)
		res.append(Series(rel))

	res = concat(res, axis=1)
	res = res.T

	if res.empty:
		pass
	
	else:
		if single_decay: 
			res.columns = ['fo_r','fm_r','tau','bias','rmse','fo_err','fm_err','tau_err','nfl','niters','flag','success']
		else:
			res.columns = ['fo_r','fm_r','alpha1','tau1','alpha2','tau2','alpha3','tau3','bias','rsme','for_err','fmr_err','alpha1_err','tau1_err','alpha2_err','tau2_err','alpha3_err','tau3_err','nfl','niters','flag','success']
		res['datetime'] = unique(dt)

	return res






#!/usr/bin/env python

from numpy import count_nonzero, isnan, inf, linalg, arange, repeat, nan
from scipy.optimize import least_squares
from sklearn import linear_model
import warnings
warnings.filterwarnings("ignore", category=RuntimeWarning)
from ._equations import __fit_kolber_nop__, __fit_kolber_p__, __fit_single_relaxation__, __fit_triple_relaxation__, __calculate_residual_saturation_p__, __calculate_residual_saturation_nop__, __calculate_residual_saturation_fixedp__, __calculate_residual_single_relaxation__, __calculate_residual_triple_relaxation__, __calculate_bias__, __calculate_rmse__, __calculate_fit_errors__
	
	
def __fit_fixed_p_model__(pfd, fyield, ro, bounds=False, sig_lims=None, method='trf', loss='soft_l1', f_scale=0.1, max_nfev=None, xtol=1e-9):

	# Count number of flashlets excluding NaNs
	nfl = count_nonzero(~isnan(fyield))
	m = ~isnan(fyield)
	fyield = fyield[m]
	pfd = pfd[m]

	# Estimates of saturation parameters
	model = linear_model.HuberRegressor()
	try:
		y = fyield[:8]
		x = arange(1,9)[:,None]
		fo_model = model.fit(x,y)
		fo = fo_model.intercept_
	except Exception:
		fo = fyield[:3].mean()
	
	try:
		y = fyield[-24:]
		x = arange(1,25)[:,None]
		fm_model = model.fit(x,y)
		fm = fm_model.intercept_
	except Exception:
		fm = fyield[-3:].mean()
	
	if (fo > fm) | (fo <= 0):
		(print('Fo greater than Fm - skipping fit.'))
		fo, fm, sigma, ro, bias, rmse, fo_err, fm_err, sigma_err, nfl, nfev = repeat(nan, 11)
		flag = -2
		success = 'False'
		return fo, fm, sigma, ro, bias, rmse, fo_err, fm_err, sigma_err, nfl, nfev, flag, success
		pass
	
	else:
		fo10 = fo * 0.1
		fm10 = fm * 0.1
		sig = 500                   
		x0 = [fo, fm, sig]

		bds = [-inf, inf]
		if bounds:
			bds = [fo-fo10, fm-fm10, sig_lims[0]], [fo+fo10, fm+fm10, sig_lims[1]]
			if (bds[0][0] > bds[1][0]) | (bds[0][1] > bds[1][1]) | (bds[0][2] > bds[1][2]):
				print('Lower bounds greater than upper bounds - fitting with no bounds.')
				bds = [-inf, inf]
		
		if max_nfev is None:
			opts = {'method':method, 'loss':loss, 'f_scale':f_scale, 'xtol':xtol} 
		else:
			opts = {'method':method, 'loss':loss, 'f_scale':f_scale, 'max_nfev':max_nfev, 'xtol':xtol} 

		try:
			popt = least_squares(__calculate_residual_saturation_fixedp__, x0, bounds=(bds), args=(pfd, fyield, ro), **opts)
			fo = popt.x[0]
			fm = popt.x[1]
			sigma = popt.x[2]

			# Calculate curve fitting statistical metrics
			sol = __fit_kolber_p__(pfd, *popt.x, ro)
			bias = __calculate_bias__(sol, fyield)
			rmse = __calculate_rmse__(popt.fun, fyield)			
			perr = __calculate_fit_errors__(popt.jac, popt.fun)
			fo_err = perr[0]
			fm_err = perr[1]
			sigma_err = perr[2]

			if max_nfev is None:
				nfev = popt.nfev
			else:
				nfev = max_nfev
			
			flag = popt.status
			success = popt.success
			
			return fo, fm, sigma, ro, bias, rmse, fo_err, fm_err, sigma_err, nfl, nfev, flag, success
		
		except linalg.LinAlgError as err:
			if str(err) == 'Singular matrix':
				print('Unable to calculate fitting errors, skipping sequence.'),
				fo, fm, sigma, ro, bias, rmse, fo_err, fm_err, sigma_err, nfl, nfev = repeat(nan, 11)
				flag = -3
				success = 'False'
				return fo, fm, sigma, ro, bias, rmse, fo_err, fm_err, sigma_err, nfl, nfev, flag, success
				pass
		except Exception:
			print('Unable to calculate fit, skipping sequence.'),
			fo, fm, sigma, ro, bias, rmse, fo_err, fm_err, sigma_err, nfl, nfev = repeat(nan, 11)
			flag = -1
			success = 'False'
			return fo, fm, sigma, ro, bias, rmse, fo_err, fm_err, sigma_err, nfl, nfev, flag, success
			pass

def __fit_calc_p_model__(pfd, fyield, bounds=False, sig_lims=None, ro_lims=None, method='trf', loss='soft_l1', f_scale=0.1, max_nfev=None, xtol=1e-9):

	# Count number of flashlets excluding NaNs
	nfl = count_nonzero(~isnan(fyield))
	m = ~isnan(fyield)
	fyield = fyield[m]
	pfd = pfd[m]
	
	# Estimates of saturation parameters
	model = linear_model.HuberRegressor()
	try:
		y = fyield[:8]
		x = arange(0,8)[:,None]
		fo_model = model.fit(x,y)
		fo = fo_model.intercept_
	except Exception:
		fo = fyield[:3].mean()
	
	try:
		y = fyield[-24:]
		x = arange(0,24)[:,None]
		fm_model = model.fit(x,y)
		fm = fm_model.intercept_
	except Exception:
		fm = fyield[-3:].mean()

	if (fo > fm) | (fo <= 0):
		(print('Fo greater than Fm - skipping fit.'))
		fo, fm, sigma, ro, bias, rmse, fo_err, fm_err, sigma_err, ro_err, nfl, nfev = repeat(nan, 12)
		flag = -2
		success = 'False'
		return fo, fm, sigma, ro, bias, rmse, fo_err, fm_err, sigma_err, ro_err, nfl, nfev, flag, success
		pass
	
	else:
		fo10 = fo * 0.1
		fm10 = fm * 0.1
		sig = 500
		ro = 0.3
		x0 = [fo, fm, sig, ro]

		bds = [-inf, inf]
		if bounds:
			bds = [fo-fo10, fm-fm10, sig_lims[0], ro_lims[0]],[fo+fo10, fm+fm10, sig_lims[1], ro_lims[1]]
			if (bds[0][0] > bds[1][0]) | (bds[0][1] > bds[1][1]) | (bds[0][2] > bds[1][2]) | (bds[0][3] > bds[1][3]): #| (bds[0][0] == 0):
				print('Lower bounds greater than upper bounds - fitting with no bounds.')
				bds = [-inf, inf]
		
		if max_nfev is None:
			opts = {'method':method, 'loss':loss, 'f_scale':f_scale, 'xtol':xtol} 
		else:
			opts = {'method':method, 'loss':loss, 'f_scale':f_scale, 'max_nfev':max_nfev, 'xtol':xtol} 

		try:
			popt = least_squares(__calculate_residual_saturation_p__, x0, bounds=(bds), args=(pfd, fyield), **opts)
			fo = popt.x[0]
			fm = popt.x[1]
			sigma = popt.x[2]
			ro = popt.x[3]

			# Calculate curve fitting statistical metrics
			sol = __fit_kolber_p__(pfd, *popt.x)
			bias = __calculate_bias__(sol, fyield)
			rmse = __calculate_rmse__(popt.fun, fyield)			
			perr = __calculate_fit_errors__(popt.jac, popt.fun)
			fo_err = perr[0]
			fm_err = perr[1]
			sigma_err = perr[2]
			ro_err = perr[3]
			
			if max_nfev is None:
				nfev = popt.nfev
			else:
				nfev = max_nfev
			
			flag = popt.status
			status = popt.success
			
			return fo, fm, sigma, ro, bias, rmse, fo_err, fm_err, sigma_err, ro_err, nfl, nfev, flag, status
		
		except linalg.LinAlgError as err:
			if str(err) == 'Singular matrix':
				print('Unable to calculate fitting errors, skipping sequence.'),
				fo, fm, sigma, ro, bias, rmse, fo_err, fm_err, sigma_err, ro_err, nfl, nfev = repeat(nan, 12)
				flag = -3
				success = 'False'
				return fo, fm, sigma, ro, bias, rmse, fo_err, fm_err, sigma_err, ro_err, nfl, nfev, flag, success
				pass

		except Exception:
			print('Unable to calculate fit, skipping sequence.'),
			fo, fm, sigma, ro, bias, rmse, fo_err, fm_err, sigma_err, ro_err, nfl, nfev = repeat(nan, 12)
			flag = -1
			success = 'False'
			return fo, fm, sigma, ro, bias, rmse, fo_err, fm_err, sigma_err, ro_err, nfl, nfev, flag, success
			pass

def __fit_no_p_model__(pfd, fyield, ro=None, bounds=False, sig_lims=None, method='trf', loss='soft_l1', f_scale=0.1, max_nfev=None, xtol=1e-9):
	
	# Count number of flashlets excluding NaNs
	nfl = count_nonzero(~isnan(fyield))
	m = ~isnan(fyield)
	fyield = fyield[m]
	pfd = pfd[m]

	# Estimates of saturation parameters
	model = linear_model.HuberRegressor()
	try:
		y = fyield[:8]
		x = arange(0,8)[:,None]
		fo_model = model.fit(x,y)
		fo = fo_model.intercept_
	except Exception:
		fo = fyield[:3].mean()
	
	try:
		y = fyield[-24:]
		x = arange(0,24)[:,None]
		fm_model = model.fit(x,y)
		fm = fm_model.intercept_
	except Exception:
		fm = fyield[-3:].mean()

	if (fo > fm) | (fo <= 0):
		(print('Fo greater than Fm - skipping fit.'))
		fo, fm, sigma, bias, rmse, fo_err, fm_err, sigma_err, nfev = repeat(nan, 9)
		flag = -2
		success = False
		return fo, fm, sigma, bias, rmse, fo_err, fm_err, sigma_err, nfl, nfev, flag, success
		pass

	else:
		fo10 = fo * 0.1
		fm10 = fm * 0.1
		sig = 500                   
		x0 = [fo, fm, sig]

		bds = [-inf, inf]
		if bounds:
			bds = [fo-fo10, fm-fm10, sig_lims[0]],[fo+fo10, fm+fm10, sig_lims[1]]
			if (bds[0][0] > bds[1][0]) | (bds[0][1] > bds[1][1]) | (bds[0][2] > bds[1][2]): #| (bds[0][0] == 0):
				print('Lower bounds greater than upper bounds - fitting with no bounds.')
				bds = [-inf, inf]

		if max_nfev is None:
			opts = {'method':method, 'loss':loss, 'f_scale':f_scale, 'xtol':xtol} 
		else:
			opts = {'method':method, 'loss':loss, 'f_scale':f_scale, 'max_nfev':max_nfev, 'xtol':xtol} 

		try:
			popt = least_squares(__calculate_residual_saturation_nop__, x0, bounds=(bds), args=(pfd, fyield), **opts)
			fo = popt.x[0] 
			fm = popt.x[1]
			sigma = popt.x[2]

			# Calculate curve fitting statistical metrics
			sol = __fit_kolber_nop__(pfd, *popt.x)
			bias = __calculate_bias__(sol, fyield)
			rmse = __calculate_rmse__(popt.fun, fyield)			
			perr = __calculate_fit_errors__(popt.jac, popt.fun)
			fo_err = perr[0]
			fm_err = perr[1]
			sigma_err = perr[2]

			if max_nfev is None:
				nfev = popt.nfev
			else:
				nfev = max_nfev
			
			flag = popt.status
			success = popt.success
			
			return fo, fm, sigma, bias, rmse, fo_err, fm_err, sigma_err, nfl, nfev, flag, success
		
		except linalg.LinAlgError as err:
			if str(err) == 'Singular matrix':
				print('Unable to calculate fitting errors, skipping sequence.'),
				fo, fm, sigma, bias, rmse, fo_err, fm_err, sigma_err, nfl, nfev = repeat(nan, 10)
				flag = -3
				success = 'False'
				return fo, fm, sigma, bias, rmse, fo_err, fm_err, sigma_err, nfl, nfev, flag, success
				pass

		except Exception:
			print('Unable to calculate fit, skipping sequence.'),
			fo, fm, sigma, bias, rmse, fo_err, fm_err, sigma_err, nfl, nfev = repeat(nan, 10)
			flag = -1
			success = 'False'
			return fo, fm, sigma, bias, rmse, fo_err, fm_err, sigma_err, nfl, nfev, flag, success
			pass

def __fit_single_decay__(seq_time, fyield, sat_flashlets=None, bounds=False, single_lims=None, method='trf', loss='soft_l1', f_scale=0.1, max_nfev=None, xtol=1e-9):
   
	# Count number of flashlets excluding NaNs
	nfl = count_nonzero(~isnan(fyield))
	m = ~isnan(fyield)
	fyield = fyield[m]
	seq_time = seq_time[m]

	# Estimates of relaxation parameters
	fo_relax = fyield[-3:].mean()
	
	if sat_flashlets is None:
		fm_relax = fyield[:3].mean()
	else:
		fm_relax = fyield[:3+sat_flashlets].mean()

	if (fo_relax > fm_relax):
		(print('Fo_relax greater than Fm_relax - skipping fit.'))
		fo_r, fm_r, tau, bias, rmse, fo_err, fm_err, tau_err, nfev = repeat(nan, 9)
		flag = -2
		success = 'False'
		return fo_r, fm_r, tau, bias, rmse, fo_err, fm_err, tau_err, nfl, nfev, flag, success
		pass

	fo10 = fo_relax * 0.1
	fm10 = fm_relax * 0.1
	tau = 4000
	x0 = [fo_relax, fm_relax, tau]

	bds = [-inf, inf]
	if bounds:
		bds = [fo_relax-fo10, fm_relax-fm10, single_lims[0]],[fo_relax+fo10, fm_relax+fm10, single_lims[1]]
		if (bds[0][0] > bds[1][0]) | (bds[0][1] > bds[1][1]) | (bds[0][2] > bds[1][2]):
			print('Lower bounds greater than upper bounds - fitting with no bounds.')
			bds = [-inf, inf]
	
	if max_nfev is None:
		opts = {'method':method, 'loss':loss, 'f_scale':f_scale, 'xtol':xtol} 
	else:
		opts = {'method':method, 'loss':loss, 'f_scale':f_scale, 'max_nfev':max_nfev, 'xtol':xtol} 

	try: 	
		popt = least_squares(__calculate_residual_single_relaxation__, x0, bounds=(bds), args=(seq_time, fyield), **opts)
		fo_r =  popt.x[0]
		fm_r = popt.x[1]
		tau = popt.x[2]

		# Calculate curve fitting statistical metrics
		sol = __fit_single_relaxation__(seq_time, *popt.x)
		bias = __calculate_bias__(sol, fyield)
		rmse = __calculate_rmse__(popt.fun, fyield)			
		perr = __calculate_fit_errors__(popt.jac, popt.fun)
		fo_err = perr[0]
		fm_err = perr[1]
		tau_err = perr[2]

		if max_nfev is None:
			nfev = popt.nfev
		else:
			nfev = max_nfev
			
		flag = popt.status
		success = popt.success

		return  fo_r, fm_r, tau, bias, rmse, fo_err, fm_err, tau_err, nfl, nfev, flag, success
	
	except linalg.LinAlgError as err:
		if str(err) == 'Singular matrix':
			print('Unable to calculate fitting errors, skipping sequence.'),
			fo_r, fm_r, tau, bias, rmse, fo_err, fm_err, tau_err, nfev = repeat(nan, 9)
			flag = -3
			success = 'False'
			return fo_r, fm_r, tau, bias, rmse, fo_err, fm_err, tau_err, nfl, nfev, flag, success
			pass
	
	except Exception:
		print('Unable to calculate fit, skipping sequence.'),
		fo_r, fm_r, tau, bias, rmse, fo_err, fm_err, tau_err, nfev = repeat(nan, 9)
		flag = -1
		success = 'False'
		return fo_r, fm_r, tau, bias, rmse, fo_err, fm_err, tau_err, nfl, nfev, flag, success
		pass


def __fit_triple_decay__(seq_time, fyield, sat_flashlets=None, bounds=False, tau1_lims=None, tau2_lims=None, tau3_lims=None, method='trf', loss='soft_l1', f_scale=0.1, max_nfev=None, xtol=1e-9):
    
	# Count number of flashlets excluding NaNs
	nfl = count_nonzero(~isnan(fyield))
	m = ~isnan(fyield)
	fyield = fyield[m]
	seq_time = seq_time[m]

	# Estimates of relaxation parameters
	fo_relax = fyield[-3:].mean()
	
	if sat_flashlets is None:
		fm_relax = fyield[:3].mean()
	else:
		fm_relax = fyield[:3+sat_flashlets].mean()

	if (fo_relax > fm_relax):
		(print('Fo_relax greater than Fm_relax - skipping fit.'))
		fo_r, fm_r, a1, t1, a2, t2, a3, t3, bias, rmse, fo_err, fm_err, a1_err, t1_err, a2_err, t2_err, a3_err, t3_err, nfl, nfev = repeat(nan, 20)
		flag = -2
		success = 'False'
		return fo_r, fm_r, a1, t1, a2, t2, a3, t3, bias, rmse, fo_err, fm_err, a1_err, t1_err, a2_err, t2_err, a3_err, t3_err, nfl, nfev, flag, success
		pass

	fo10 = fo_relax * 0.1
	fm10 = fm_relax * 0.1
	alpha1 = 0.3
	tau1 = 600
	alpha2 = 0.3
	tau2 = 2000
	alpha3 = 0.3
	tau3 = 30000
	x0 = [fo_relax, fm_relax, alpha1, tau1, alpha2, tau2, alpha3, tau3]
    
	bds = [-inf, inf]
	if bounds:
		bds = [fo_relax-fo10, fm_relax-fm10, 0.1, tau1_lims[0], 0.1, tau2_lims[0], 0.1, tau3_lims[0]],[fo_relax+fo10, fm_relax+fm10, 1, tau1_lims[1], 1, tau2_lims[1], 1, tau3_lims[1]]
		if (bds[0][0] > bds[1][0]) | (bds[0][1] > bds[1][1]) | (bds[0][2] > bds[1][2]) | (bds[0][3] > bds[1][3]) | (bds[0][4] > bds[1][4]) | (bds[0][5] > bds[1][5]) | (bds[0][6] > bds[1][6]) | (bds[0][7] > bds[1][7]):
			print('Lower bounds greater than upper bounds - fitting with no bounds.')
			bds = [-inf, inf]
	
	if max_nfev is None:
		opts = {'method':method, 'loss':loss, 'f_scale':f_scale, 'xtol':xtol} 
	else:
		opts = {'method':method, 'loss':loss, 'f_scale':f_scale, 'max_nfev':max_nfev, 'xtol':xtol} 

	try: 	
		popt = least_squares(__calculate_residual_triple_relaxation__, x0, bounds=(bds), args=(seq_time, fyield), **opts)
		fo_r =  popt.x[0]
		fm_r = popt.x[1]
		a1 = popt.x[2]
		t1 = popt.x[3]
		a2 = popt.x[4]
		t2 = popt.x[5]
		a3 = popt.x[6]
		t3 = popt.x[7]

		# Calculate curve fitting statistical metrics
		sol = __fit_single_relaxation__(seq_time, *popt.x)
		bias = __calculate_bias__(sol, fyield)
		rmse = __calculate_rmse__(popt.fun, fyield)			
		perr = __calculate_fit_errors__(popt.jac, popt.fun)
		fo_err = perr[0]
		fm_err = perr[1]
		a1_err = perr[2]
		t1_err = perr[3]
		a2_err = perr[4]
		t2_err = perr[5]
		a3_err = perr[6]
		t3_err = perr[7]

		if max_nfev is None:
			nfev = popt.nfev
		else:
			nfev = max_nfev
		
		flag = popt.status
		success = popt.success

		return  fo_r, fm_r, a1, t1, a2, t2, a3, t3, bias, rmse, fo_err, fm_err, a1_err, t1_err, a2_err, t2_err, a3_err, t3_err, nfl, nfev, flag, success
	
	except linalg.LinAlgError as err:
		if str(err) == 'Singular matrix':
			print('Unable to calculate fitting errors, skipping sequence.'),
			fo_r, fm_r, a1, t1, a2, t2, a3, t3, bias, rmse, fo_err, fm_err, a1_err, t1_err, a2_err, t2_err, a3_err, t3_err, nfl, nfev = repeat(nan, 20)
			flag = -3
			success = 'False'
			return fo_r, fm_r, a1, t1, a2, t2, a3, t3, bias, rmse, fo_err, fm_err, a1_err, t1_err, a2_err, t2_err, a3_err, t3_err, nfl, nfev, flag, success
			pass
	
	except Exception:
		print('Unable to calculate fit, skipping sequence.'),
		fo_r, fm_r, a1, t1, a2, t2, a3, t3, bias, rmse, fo_err, fm_err, a1_err, t1_err, a2_err, t2_err, a3_err, t3_err, nfl, nfev = repeat(nan, 20)
		flag = -1
		success = 'False'
		return fo_r, fm_r, a1, t1, a2, t2, a3, t3, bias, rmse, fo_err, fm_err, a1_err, t1_err, a2_err, t2_err, a3_err, t3_err, nfl, nfev, flag, success
		pass




#!/usr/bin/env python

from ._equations import __fit_kolber_p__, __fit_kolber_nop__, __fit_single_relaxation__, __fit_triple_relaxation__, __calculate_Webb_model__, __calculate_modified_Webb_model__
from matplotlib.pyplot import subplots, close
from numpy import arange, array

def plot_saturation_data(fyield, pfd, fo=None, fm=None, sigma=None, ro=None, rmse=None):

	"""
	Parameters
	----------

	fyield : np.array, dtype=float, shape=[n,]
		The raw fluorescence yield data.
	pfd : np.array, dtype=float, shape=[n,]
		The photon flux density.
	fo : float, default=None
		The minimum fluorescence value.
	fm : float, default=None
		The maximum fluorescence value.
	sigma: float, default=None
		The effective absorption cross-section value in Å\ :sup:`2`.
	ro: float, default=None
		The connectivity coefficient.
	rmse: float, default=None
		The RMSE value of the fit.

	Returns
	-------

	ax : object
		A matplotlib figure object.

	Example
	-------
	>>> plot_saturation_data(fyield, pfd, fo=fo, fm=fm, sigma=sigma, ro=None, rmse=rmse)
	"""

	fyield = array(fyield)
	pfd = array(pfd)
	fvfm = (fm - fo)/fm
	x = arange(0,len(fyield),1)

	close()

	fig, ax = subplots(1, 1, figsize=[5,4], dpi=90)

	ax.plot(x, fyield, marker='o', lw=0, label='Raw Data', color='0.5')
	ax.set_ylabel('Fluorescence Yield')
	ax.set_xlabel('Flashlet Number')

	if ro is None:
		params = [fo, fm, sigma]
		formula = r"F$_v$/F$_m$ = {:.2f}""\n""$\u03C3$$_{{PSII}}$ = {:.2f}; RMSE = {:.2f}".format(fvfm, sigma, rmse)
		ax.plot(x, __fit_kolber_nop__(pfd, *params), color='k', label='{}'.format(formula))
	else:
		params = [fo, fm, sigma, ro]
		formula = r"F$_v$/F$_m$ = {:.2f}""\n""$\u03C3$$_{{PSII}}$ = {:.2f}; RMSE = {:.2f}".format(fvfm, sigma, rmse)
		ax.plot(x, __fit_kolber_p__(pfd, *params), color='k', label='{}'.format(formula))

	ax.legend()
	

	return ax

def plot_relaxation_data(fyield, seq_time, fo_relax=None, fm_relax=None, tau=None, alpha=None, rmse=None):
	"""
	Parameters
	----------

	fyield : np.array, dtype=float, shape=[n,]
		The raw fluorescence yield data.
	seq_time : np.array, dtype=float, shape=[n,]
		The time of the flashlet measurements.
	fo_relax : float, default=None
		The minimum fluorescence value in the relaxation phase.
	fm_relax : float, default=None
		The maximum fluorescence value in the relaxation phase.
	tau: float, default=None
		The rate of reoxidation in μs.
	alpha: float, default=None
		The ratio of reoxidisation components.
	rmse: float, default=None
		The RMSE value of the fit.

	Returns
	-------

	ax : object
		A matplotlib figure object.
	
	Example
	-------
	>>> ppu.plot_relaxation_data(fyield, seq_time, fo_relax=fo_r, fm_relax=fm_r, tau=(tau1, tau2, tau3), alpha=(alpha1, alpha2, alpha3), rsq=rsq)
	"""

	fyield = array(fyield)
	seq_time = array(seq_time)

	x = arange(0,len(fyield),1)+100

	close()

	fig, ax = subplots(1, 1, figsize=[5,4], dpi=90)

	ax.plot(x, fyield, marker='o', lw=0, label='Raw Data', color='0.5')
	ax.set_ylabel('Fluorescence Yield')
	ax.set_xlabel('Flashlet Number')

	if alpha is None:
		params = [fo_relax, fm_relax, tau]
	
		formula = r"F$_o$$_{{Relax}}$ = {:.2f}; F$_m$$_{{Relax}}$ = {:.2f}""\n""$\U0001D70F$ = {:.2f}; RMSE = {:.2f}".format(fo_relax, fm_relax, tau, rmse)
		ax.plot(x, __fit_single_relaxation__(seq_time, *params), color='k', label='{}'.format(formula))

	else:
		params = [fo_relax, fm_relax, alpha[0], tau[0], alpha[1], tau[1], alpha[2], tau[2]]

		formula = r"F$_o$$_{{Relax}}$ = {:.2f}; F$_m$$_{{Relax}}$ = {:.2f}""\n""$\U0001D70F$$_1$ = {:.2f}; $\U0001D70F$$_2$ = {:.2f}""\n""$\U0001D70F$$_3$ = {:.2f}; RMSE = {:.2f}".format(fo_relax, fm_relax, tau[0], tau[1], tau[2], rmse)
		ax.plot(x, __fit_triple_relaxation__(seq_time, *params), color='k', label='{}'.format(formula))

	ax.legend()


	return ax


def plot_fluorescence_light_curve(par, etr, etrmax=None, alpha=None, rmse=None, sigma=None, phi=False):
	
	"""
	Parameters
	----------

	par : np.array, dtype=float
		The actinic light data from the fluorescence light curve.
	etr : np.array, dtype=float
		The electron transport rate data.
	etrmax : float, default=None
		The maximum electron transport rate.
	alpha : float, default=None
		The light limited slope of electron transport.
	rmse: float, default=None
		The RMSE value of the fit.
	sigma: float, default=None
		The effective absorption-cross section.
	phi: bool, default=False
		If True, etr data is phi and the modified Webb et al. (1974) fit is used.

	Returns
	-------

	ax : object
		A matplotlib figure object.

	Example
	-------
	>>> ppu.plot_fluorescence_light_curve(par, etr, etrmax=etr_max, alpha=alpha, rsq=rsq, sigma=sigma, phi=True)
	"""

	x = array(par)
	y = array(etr)
	
	close()
	fig, ax = subplots(1, 1, figsize=[5,4], dpi=90)

	ax.plot(x, y, marker='o', lw=0, label='Raw Data', color='0.5')
	formula = r"ETR$_{{max}}$ = {:.2f}""\n""$\u03B1$$^{{ETR}}$ = {:.2f}""\n"" RMSE = {:.2f}".format(etrmax, alpha, rmse)
	ax.set_xlabel('Actinic Light ($\u03BC$mol photons m$^{-2}$ s${-1}$)')
	
	if phi == False:
		params = [etrmax, alpha]
		ax.plot(x, __calculate_Webb_model__(x, *params), color='k', label='{}'.format(formula))
		ax.set_ylabel('ETR (mol e$^{-1}$ mol RCII$^{-1}$ s$^{-1}$)')
	
	else:
		if sigma is None:
			print('UserError - no sigma data provided.')
		sig = sigma*6.022e-3
		params = [etrmax/sig, alpha/sig]
		ax.plot(x, __calculate_modified_Webb_model__(x, *params), color='k', label='{}'.format(formula))
		ax.set_xscale('log')
		ax.set_ylabel('\u03D5')
	
	ax.legend()

	
	return ax



from setuptools import setup, find_packages
import os


def read(fname):
    fpath = os.path.join(os.path.dirname(__file__), fname)
    return open(fpath).read()


def find_version_from_readme():
    s = read('README.md')
    i0 = s.lower().find('version')
    i1 = i0 + 20
    v = s[i0:i1].splitlines()[0]  # removes next line
    v = v.split(' ')[1]  # finds version number
    return v


def walker(base, *paths):
    file_list = set([])
    cur_dir = os.path.abspath(os.curdir)

    os.chdir(base)
    try:
        for path in paths:
            for dname, _, files in os.walk(path):
                for f in files:
                    file_list.add(os.path.join(dname, f))
    finally:
        os.chdir(cur_dir)

    return list(file_list)


setup(
    # Application name:
    name='phyto_photo_utils',

    # Version number (initial):
    version=find_version_from_readme(),

    # Application author details:
    author='Thomas Ryan-Keogh, Charlotte Robinson',
    author_email='tjryankeogh@gmail.com',

    # Packages
    packages=find_packages(),

    # Include additional files into the package
    include_package_data=True,

    # Details
    url='https://gitlab.com/tjryankeogh/phytophotoutils',
    download_url= 'https://gitlab.com/tjryankeogh/phytophotoutils/-/archive/v1.3/phytophotoutils-v1.3.tar.gz',
    license='MIT License',
    description='Tools and utilities for active chlorophyll fluorescence data processing.',
    organisation='Council for Scientific and Industrial Research, Curtin University',

    long_description=read('README.md'),
    long_description_content_type='text/markdown',

    # Dependent packages (distributions)
    install_requires=[
        'tqdm',
        'scipy',
        'numpy',
        'pandas',
        'datetime',
        'matplotlib',
        'sklearn'],
)



#!/usr/bin/env python
	
from ._fitting import __fit_fixed_p_model__, __fit_calc_p_model__, __fit_no_p_model__
from pandas import Series, concat
from numpy import array, unique
from tqdm import tqdm


def fit_saturation(pfd, fyield, seq, datetime, blank=0, sat_len=100, skip=0, ro=0.3, no_ro=False, fixed_ro=False, bounds=True, sig_lims =[100, 2200], ro_lims=[0.0, 1.0], method='trf', loss='soft_l1', f_scale=0.1, max_nfev=None, xtol=1e-9):
    
	"""
	Process the raw transient data and perform the Kolber et al. 1998 saturation model.


	Parameters
	----------
	pfd : np.array, dtype=float, shape=[n,] 
		The photon flux density of the instrument in μmol photons m\ :sup:`2` s\ :sup:`-1`.
	fyield : np.array, dtype=float, shape=[n,] 
		The fluorescence yield of the instrument.
	seq : np.array, dtype=int, shape=[n,] 
		The measurement number.
	datetime : np.array, dtype=datetime64, shape=[n,]
		The date & time of each measurement.
	blank : np.array, dype=float, shape=[n,]
		The blank value, must be the same length as fyield.
	sat_len : int, default=100
		The number of flashlets in saturation sequence.
	skip : int, default=0
		the number of flashlets to skip at start.
	ro : float, default=0.3
		The fixed value of the connectivity coefficient. Not required if fixed_ro is False.
	no_ro : bool, default=False
		If True, this processes the raw transient data and performs the no connectivity saturation model.
	fixed_ro : bool, default=False
		If True, this sets a user defined fixed value for ro (the connectivity factor) when fitting the saturation model.
	bounds : bool, default=True
		If True, will set lower and upper limit bounds for the estimation, not suitable for methods 'lm'.
	sig_lims : [int, int], default=[100, 2200]
	 	The lower and upper limit bounds for fitting sigmaPSII.
	ro_lims: [float, float], default=[0.0, 0.1]
		The lower and upper limit bounds for fitting the connectivity coefficient. Not required if no_ro and fixed_ro are False.
	method : str, default='trf'
		The algorithm to perform minimization. See ``scipy.optimize.least_squares`` documentation for more information on non-linear least squares fitting options.
	loss : str, default='soft_l1'
		The loss function to be used. Note: Method ‘lm’ supports only ‘linear’ loss. See ``scipy.optimize.least_squares`` documentation for more information on non-linear least squares fitting options.
	f_scale : float, default=0.1
	 	The soft margin value between inlier and outlier residuals. See ``scipy.optimize.least_squares`` documentation for more information on non-linear least squares fitting options.
	max_nfev : int, default=None		
		The number of iterations to perform fitting routine. If None, the value is chosen automatically. See ``scipy.optimize.least_squares`` documentation for more information on non-linear least squares fitting options.
	xtol : float, default=1e-9			
		The tolerance for termination by the change of the independent variables. See ``scipy.optimize.least_squares`` documentation for more information on non-linear least squares fitting options.

	Returns
	-------

	res: pandas.DataFrame
		The results of the fitting routine with columns as below:
	fo : np.array, dtype=float, shape=[n,]
		The minimum fluorescence level.
	fm : np.array, dtype=float, shape=[n,]
		The maximum fluorescence level.
	sigma : np.array, dtype=float, shape=[n,]
		The effective absorption cross-section of PSII in Å\ :sup:`2`.
	fvfm : np.array, dtype=float, shape=[n,]
		The maximum photochemical efficiency.
	ro : np.array, dtype=float, shape=[n,]
		The connectivity coefficient, ρ, only returned if no_ro and fixed_ro are False.
	bias : np.array, dtype=float, shape=[n,]
		The bias of fit.
	rmse : np.array, dtype=float, shape=[n,]
		The root mean squared error of the fit.
	fo_err : np.array, dtype=float, shape=[n,]
		The fit error of F\ :sub:`o`.
	fm_err : np.array, dtype=float, shape=[n,]
		The fit error of F\ :sub:`m`.
	sigma_err : np.array, dtype=float, shape=[n,]
		The fit error of σ\ :sub:`PSII`.
	ro_err : np.array, dtype=float, shape=[n,]
		The fit error of ρ, only returned if no_ro and fixed_ro are False.
	nfl : np.array, dtype=int, shape=[n,]
		The number of flashlets used for fitting.
	niters : np.array, dype=int, shape=[n,]
		The number of functional evaluations done on the fitting routine.
	flag : np.array, dtype=int, shape=[n,]
		The code associated with the fitting routine success, positive values = SUCCESS, negative values = FAILURE.
		-3 : Unable to calculate parameter errors
		-2 : F\ :sub:`o` > F\ :sub:`m`
		-1 : improper input parameters status returned from MINPACK.
		0 : the maximum number of function evaluations is exceeded.
		1 : gtol termination condition is satisfied.
		2 : ftol termination condition is satisfied.
		3 : xtol termination condition is satisfied.
		4 : Both ftol and xtol termination conditions are satisfied.
	success : np.array, dtype=bool, shape=[n,]
		A boolean array reporting whether fit was successful (TRUE) or if not successful (FALSE)
	datetime : np.array, dtype=datetime64, shape=[n,]
		The date and time associated with the measurement.

	Example
	-------
	>>> sat = ppu.calculate_saturation(pfd, fyield, seq, datetime, blank=0, sat_len=100, skip=0, ro=0.3, no_ro=False, fixed_ro=True, sig_lims =[100,2200])
	"""

	pfd = array(pfd)
	fyield = array(fyield)
	seq = array(seq)
	dt = array(datetime)
	
	res = []
	
	if no_ro:
		opts = {'bounds':bounds, 'sig_lims':sig_lims, 'method':method,'loss':loss, 'f_scale':f_scale, 'max_nfev':max_nfev, 'xtol':xtol} 
	elif fixed_ro:
		opts = {'bounds':bounds, 'sig_lims':sig_lims, 'method':method,'loss':loss, 'f_scale':f_scale, 'max_nfev':max_nfev, 'xtol':xtol} 
	else:
		opts = {'bounds':bounds, 'sig_lims':sig_lims, 'ro_lims':ro_lims, 'method':method,'loss':loss, 'f_scale':f_scale, 'max_nfev':max_nfev, 'xtol':xtol} 
	
	for s in tqdm(unique(seq)):
	    
		i = seq == s
		x = pfd[i]
		y = fyield[i]
		if no_ro:
			sat = __fit_no_p_model__(x[skip:sat_len], y[skip:sat_len], **opts)
		elif fixed_ro:
			sat = __fit_fixed_p_model__(x[skip:sat_len], y[skip:sat_len], ro, **opts)
		else:
			sat = __fit_calc_p_model__(x[skip:sat_len], y[skip:sat_len], **opts)

		res.append(Series(sat))

	res = concat(res, axis=1)
	res = res.T

	if res.empty:
		pass
	
	else: 
		if no_ro:
			res.columns = ['fo','fm','sigma','bias','rmse','fo_err','fm_err','sigma_err','nfl','niters','flag','success']
		if fixed_ro:
			res.columns = ['fo','fm','sigma','ro','bias','rmse','fo_err','fm_err','sigma_err','nfl','niters','flag','success']
		else:
			res.columns = ['fo','fm','sigma','ro','bias','rmse','fo_err','fm_err','sigma_err','ro_err','nfl','niters','flag','success']
	
		# Subtract blank from Fo and Fm 
		res['fo'] -= blank
		res['fm'] -= blank

		# Calculate Fv/Fm after blank subtraction
		fvfm = (res.fm - res.fo)/res.fm
		res.insert(3, "fvfm", fvfm)
		
		res['datetime'] = unique(dt)
	
	return res


#!/usr/bin/env python

from ._equations import __calculate_residual_etr__, __calculate_residual_phi__, __calculate_Webb_model__, __calculate_modified_Webb_model__, __calculate_bias__, __calculate_fit_errors__, __calculate_rmse__
from numpy import mean, array, isnan, inf, repeat, nan, concatenate
from pandas import DataFrame
from scipy.optimize import least_squares
import warnings

def calculate_etr(fo, fm, sigma, par, light_independent=True, dark_sigma=False, light_step_size=None, outlier_multiplier=3, return_data=False, bounds=True, alpha_lims=[0,4], etrmax_lims=[0,2000], method='trf', loss='soft_l1', f_scale=0.1, max_nfev=None, xtol=1e-9):
      
	"""
	
	Convert the processed transient data into an electron transport rate and perform a fit using the Webb Model.

	Parameters
	----------
	fo : np.array, dtype=float, shape=[n,]
		The minimum fluorescence level.
	fm : np.array, dtype=float, shape=[n,] 
		The maximum fluorescence level.
	sigma : np.array, dtype=float, shape=[n,] 
		The effective absorption cross-section of PSII in Å\ :sup:`2`.
	par : np.array, dtype=float, shape=[n,]
		The actinic light levels in μE m\ :sup:`2` s\ :sup:`-1`.
	light_independent : bool, default=True
		If True, will use the method outlined in Silsbe & Kromkamp 2012. 
	dark_sigma : bool
		If True, will use mean of σ\ :sub:`PSII` under 0 actinic light for calculation. If False, will use σ\ :sub:`PSII` and σ\ :sub:`PSII`' for calculation.
	light_step_size : int
		The number of measurements for initial light step.
	outlier_multiplier : int, default=3
		The multiplier to apply to the standard deviation for determining the upper and lower limits.
	return_data : bool, default=False
		If True, will return the final data used for the fit.
	bounds : bool, default=True
		If True, will set lower and upper limit bounds for the estimation, not suitable for methods 'lm'.
	alpha_lims : [int, int], default=[0,4]
		The lower and upper limit bounds for fitting α\ :sup:`ETR`.
	etrmax_lims : [int, int], default=[0,2000]
	 	The lower and upper limit bounds for fitting ETR\ :sub:`max`.
	fit_method : str, default='trf'
		The algorithm to perform minimization. See ``scipy.optimize.least_squares`` documentation for more information on non-linear least squares fitting options.
	loss_method : str, default='soft_l1'
		The loss function to be used. Note: Method ‘lm’ supports only ‘linear’ loss. See ``scipy.optimize.least_squares`` documentation for more information on non-linear least squares fitting options.
	fscale : float, default=0.1
	 	The soft margin value between inlier and outlier residuals. See ``scipy.optimize.least_squares`` documentation for more information on non-linear least squares fitting options.
	max_nfev : int, default=None		
		The number of iterations to perform fitting routine. If None, the value is chosen automatically. See ``scipy.optimize.least_squares`` documentation for more information on non-linear least squares fitting options.
	xtol : float, default=1e-9			
		The tolerance for termination by the change of the independent variables. See ``scipy.optimize.least_squares`` documentation for more information on non-linear least squares fitting options.

	Returns
	-------
	
	etr_max : float
		The maximum electron transport rate.
	alpha : float
		The light limited slope of electron transport.
	ek : float
		The photoacclimation of ETR.
	rsq : np.array, dtype=float, shape=[n,]
		The r\ :sup:`2` value.
	bias : np.array, dtype=float, shape=[n,]
		The bias of fit.
	chi : np.array, dtype=float, shape=[n,]
		The chi-squared goodness of fit.
	rchi : np.array, dtype=float, shape=[n,]
		The reduced chi-squared goodness of fit.
	rmse : np.array, dtype=float, shape=[n,]
		The root mean squared error of the fit.
	etrmax_err : float
		The fit error of ETR\ :sup:`max`.
	alpha_err : float
		The fit error of α\ :sub:`ETR`.
	data : [np.array, np.array]
		Optional, the final data used for the fitting procedure.


	Example
	-------
	>>> etr_max, alpha, ek, rsq, bias, chi, etr_max_err, alpha_err = ppu.calculate_e_dependent_etr(fo, fm, fvfm, sigma, par, return_data=False)
	"""

	warnings.simplefilter(action = "ignore", category = RuntimeWarning)

	fo = array(fo)
	fm = array(fm)
	fvfm = (fm - fo) / fm
	sigma = array(sigma)
	par = array(par)

	lss = light_step_size - 1 # Python starts at 0
	
	if light_independent:
		etr = fvfm / mean(fvfm[0:lss])
	else:
		if dark_sigma:
			etr = (par * mean(sigma[0:lss]) * (fvfm / mean(fvfm[0:lss]))) * 6.022e-3
	
		else:
			f_o = mean(fo[0:lss]) / (mean(fvfm[0:lss]) + (mean(fo[0:lss])/fm))
			fqfv = (fm - fo) / (fm - f_o)
			etr = par * sigma * fqfv * 6.022e-3

	df = DataFrame([par, etr])
	df = df.T
	df.columns = ['par', 'etr']

	# exclude outliers if more than mean ± (stdev * multiplier)
	grp = df.groupby(by='par')
	mn = grp.mean()
	std = grp.std()
	c = grp.count()
	ulim = repeat((mn.etr.values + std.etr.values * outlier_multiplier), c.etr.values)
	llim = repeat((mn.etr.values - std.etr.values * outlier_multiplier), c.etr.values)
	idx = []
	for i, items in enumerate(grp.indices.items()):
		idx.append(items[-1])

	idx = concatenate(idx, axis=0)

	# Create pandas DataFrame of upper and lower using original indexes of data
	mask = DataFrame([ulim, llim, idx]).T
	mask.columns = ['ulim','llim','index']
	mask = mask.set_index('index').sort_index()

	m = (df.etr.values > mask.ulim) | (df.etr.values < mask.llim)

	# Where condition is True, set values of value to NaN
	df.loc[m.values,'etr'] = nan

	# Create means per light step
	df = df.groupby('par').mean().reset_index()
	#TO DO apply function of excluding outliers from means

	# Define data for fitting and estimates of ETRmax and alpha
	P = array(df.etr)
	E = array(df.par)

	p0 = [1000, 1.5]

	# Mask missing data
	if light_independent:
		mask = isnan(P) | isnan(E) | (P < 0) | (E == 0)
	else:
		mask = isnan(P) | isnan(E)
	
	E = E[~mask]
	P = P[~mask]
	
	if bounds:
		bds = [etrmax_lims[0], alpha_lims[0]],[etrmax_lims[1], alpha_lims[1]]
		if (bds[0][0] > bds[1][0]) | (bds[0][1] > bds[1][1]):
			print('Lower bounds greater than upper bounds - fitting with no bounds.')
			bds = [-inf, inf]
	else:
		bds = [-inf, inf]

	if max_nfev is None:
		opts = {'method':method, 'loss':loss, 'f_scale':f_scale, 'xtol':xtol} 
	else:
		opts = {'method':method, 'loss':loss, 'f_scale':f_scale, 'max_nfev':max_nfev, 'xtol':xtol} 

	try:
		if light_independent:
			popt = least_squares(__calculate_residual_phi__, p0, args=(E, P), bounds=(bds), **opts)
		else:
			popt = least_squares(__calculate_residual_etr__, p0, args=(E, P), bounds=(bds), **opts)

		if light_independent:
			etr_max = popt.x[0]
			alpha = popt.x[1]
			sigma = mean(sigma[0:lss])
			etr_max *= sigma * 6.022e-3
			alpha *= sigma * 6.022e-3
		else:
			etr_max = popt.x[0]
			alpha = popt.x[1]

		ek = etr_max / alpha
		
		if light_independent:
			sol = __calculate_modified_Webb_model__(E, *popt.x)
		else:
			sol = __calculate_Webb_model__(E, *popt.x)

		bias = __calculate_bias__(sol, P)
		rmse = __calculate_rmse__(popt.fun, P)				
		perr = __calculate_fit_errors__(popt.jac, popt.fun)
		etr_max_err = perr[0]
		alpha_err = perr[1]
	
	except Exception:
		print(('Unable to calculate fit, skipping sequence'))
		etr_max, alpha, ek, bias, rmse, etr_max_err, alpha_err = repeat(nan, 7)
	
	if return_data:
		return etr_max, alpha, ek, bias, rmse, etr_max_err, alpha_err, [E,P]
	else:
		return etr_max, alpha, ek, bias, rmse, etr_max_err, alpha_err
